"""Objects that represent FIT file developer message fields."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


from .fields import NamedField
from .object_fields import ObjectField
from .measurement import Distance, Speed


class DevField(NamedField):
    """Class that handles a developer fields."""

    def __init__(self, name, units, scale=None, offset=None, **kwargs):
        """Return a DevField instance."""
        args = {
            'name'  : name,
            'units' : units
        }
        if scale is not None:
            args['scale'] = scale
        if offset is not None:
            args['offset'] = offset
        args.update(kwargs)
        super().__init__(**args)


class DevObjectField(ObjectField):
    """Field held in a measurement object."""

    def __init__(self, name, scale, offset, obj_func, output_func):
        """Return a new instance of DevSpeedField."""
        args = {
            'name'  : name,
        }
        if scale is not None:
            scale['scale'] = 1.0
        if offset is not None:
            offset['offset'] = 0.0
        super().__init__(obj_func, output_func, **args)


class DevDistanceField(DevObjectField):
    """Field holding a distance."""

    def __init__(self, name, units, scale, offset, output_func=Distance.feet_or_meters):
        """Return a new instance of DevDistanceField."""
        super().__init__(name, scale, offset, Distance.from_func(units), output_func)


class DerivedDevDistanceField(DevDistanceField):
    """Field holding a distance."""

    def __init__(self, name, units, scale, offset, object_field):
        """Return a new instance of DerivedDevDistanceField."""
        super().__init__(name, units, scale, offset, object_field.output_func)


class DevSpeedField(DevObjectField):
    """Field holding a speed."""

    def __init__(self, name, units, scale, offset, output_func=Speed.mph_or_kph):
        """Return a new instance of DevSpeedField."""
        super().__init__(name, scale, offset, Speed.from_func(units), output_func)


class DerivedDevSpeedField(DevSpeedField):
    """Field holding a speed."""

    def __init__(self, name, units, scale, offset, object_field):
        """Return a new instance of DerivedDevSpeedField."""
        super().__init__(name, units, scale, offset, object_field.output_func)



"""Units strings for measurement systems."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import enum

from .field_enums import DisplayMeasure


class UnitTypes(enum.Enum):
    """An enumeration of unit types."""

    distance_long   = 1
    pace            = 2
    speed           = 3
    altitude        = 4
    tempurature     = 5


unit_strings = {
    DisplayMeasure.statute : {
        UnitTypes.distance_long : 'miles',
        UnitTypes.pace          : 'per mile',
        UnitTypes.speed         : 'mph',
        UnitTypes.altitude      : 'feet',
        UnitTypes.tempurature   : 'F'
    },
    DisplayMeasure.metric : {
        UnitTypes.distance_long : 'kilometers',
        UnitTypes.pace          : 'per kilometers',
        UnitTypes.speed         : 'kph',
        UnitTypes.altitude      : 'meters',
        UnitTypes.tempurature   : 'C'
    }
}



"""Object that represents a FIT field message field value."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


class FieldValue(dict):
    """Object that represents a FIT field message field value."""

    def __init__(self, field, orig, invalid, **kwargs):
        """Return a FieldValue instance given the field."""
        self.field = field
        self.orig = orig
        self.invalid = invalid
        self.update(kwargs)

    def first(self):
        """Return the first value. Most of the time there is only one."""
        return next(iter(self.values()))

    def is_invalid(self):
        """Return if the field value is valid."""
        return self.field.is_invalid(self.orig, self.invalid)

    def reconvert(self, measurement_system):
        """Redo field conversion based on new information."""
        self.update(self.field.reconvert(self.orig, self.invalid, measurement_system))

    def __str__(self):
        """Return a string representation of a FieldValue instance."""
        field_string = ""
        for field_name, value in self.items():
            field_string = field_name + "("
            if self.is_invalid():
                field_string += "[invalid]"
            else:
                field_string += repr(value)
            if self.field.units:
                field_string += " %s" % self.field.units
            if value != self.orig:
                field_string += " (%r)" % self.orig
            field_string += ")"
        return field_string

    def __repr__(self):
        """Return a string representation of a FieldValue instance."""
        return str(self)



"""FIT file data field."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


import collections

from .data import Data, Schema


class DataField(Data):
    """FIT file data field."""

    field_value = None
    __schema_cache = {}

    def __init__(self, file, definition_message, field_definition, measurement_system):
        """Return an instance of the DataField class."""
        self.field_definition = field_definition
        self.measurement_system = measurement_system
        self.field = definition_message.field(field_definition.field_definition_number)
        schema = self.__get_schema(field_definition.type_string(), field_definition.type_count())
        super().__init__(file, schema, None, definition_message.endian)

    def __populate_schema_cache(self, schema_sig, type, count):
        """Cache schema on the assumption that the set of schemas is much smaller than the number of times they are used."""
        schema = Schema(schema_sig, collections.OrderedDict([('field_value', [type, count])]))
        self.__schema_cache[schema_sig] = schema
        return schema

    def __get_schema(self, type, count):
        schema_sig = f'{type}_{count}'
        if schema_sig in self.__schema_cache:
            return self.__schema_cache[schema_sig]
        return self.__populate_schema_cache(schema_sig, type, count)

    def _convert(self):
        self.values = self.field.convert(self.field_value, self.field_definition.invalid(), self.measurement_system)

    def __str__(self):
        """Return a string reprsentation of the DataField instance."""
        return f'<DataField: {self.values}'



"""Enums that represent FIT file message product field values."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


import re

from .field_enums import FieldEnum, UnknownEnumValue
from .manufacturer import Manufacturer


class MainDeviceType(FieldEnum):
    """Device type for the main device."""

    fitness_tracker = 0
    bike_gps        = 1
    standalone_gps  = 2
    software        = 3
    other           = 254
    invalid         = 255

    @classmethod
    def derive_device_type(cls, manufacturer, product):
        """Return a device type for the device inferred from its manufactuer and product information."""
        device_type_mappings = {
            Manufacturer.Garmin: {
                cls.fitness_tracker  : r'Vivo|Forerunner|Fenix',
                cls.bike_gps         : r'Edge',
                cls.standalone_gps   : r'GPSMap',
                cls.software         : r'connect|Training_Center'
            }
        }
        if manufacturer in device_type_mappings:
            for device_type, regex in device_type_mappings[manufacturer].items():
                if re.search(regex, product.name, re.IGNORECASE):
                    return device_type
        return cls.other


class AntplusDeviceType(FieldEnum):
    """Device types for devices connected to the main device via ANT+. Like a external heart rate monitor."""

    antfs                       = 1
    bike_power                  = 11
    environment_sensor_legacy   = 12
    multi_sport_speed_distance  = 15
    control                     = 16
    fitness_equipment           = 17
    blood_pressure              = 18
    geocache_node               = 19
    light_electric_vehicle      = 20
    env_sensor                  = 25
    racquet                     = 26
    control_hub                 = 27
    run                         = 30
    muscle_oxygen               = 31
    bike_light_main             = 35
    bike_light_shared           = 36
    exd                         = 38
    bike_radar                  = 40
    weight_scale                = 119
    heart_rate                  = 120
    bike_speed_cadence          = 121
    bike_cadence                = 122
    bike_speed                  = 123
    stride_speed_distance       = 124


class LocalDeviceType(FieldEnum):
    """Device types for sub-devices resident on the main device. Like sensors embedded on a watch."""

    gps                             = 0
    accelerometer                   = 3
    barometer                       = 4
    bluetooth_low_energy_chipset    = 8
    wrist_heart_rate                = 10
    sensor_hub                      = 12
    invalid                         = 255


class UnknownDeviceType(UnknownEnumValue):
    """Device type enums for device types that we don't know how to interpret."""



"""Enums that represent FIT file message product field values."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

from .field_enums import FuzzyFieldEnum, UnknownEnumValue
from .manufacturer import Manufacturer


class GarminProduct(FuzzyFieldEnum):
    """Garmin product codes used in FIT files."""

    HRM1                            = 1
    axh01                           = 2
    axb01                           = 3
    axb02                           = 4
    hrm2ss                          = 5
    dsi_alf02                       = 6
    hrm3ss                          = 7
    hrm_run_single_byte_product_id  = 8
    Bike_Speed_Sensor               = 9
    Bike_Cadence_Sensor             = 10
    axs01                           = 11
    hrm_tri_single_byte_product_id  = 12
    fr225_single_byte_product_id    = 14
    Bike_Speed_Sensor_Gen3_sb       = 15
    Bike_Cadence_Sensor_Gen3_sb     = 16
    fr301_china                     = 473
    fr301_japan                     = 474
    fr301_korea                     = 475
    fr301_taiwan                    = 494
    fr405                           = 717
    fr50                            = 782
    fr405_japan                     = 987
    fr60                            = 988
    dsi_alf01                       = 1011
    fr310xt                         = 1018
    edge500                         = 1036
    fr110                           = 1124
    edge800                         = 1169
    edge500_taiwan                  = 1199
    edge500_japan                   = 1213
    chirp                           = 1253
    fr110_japan                     = 1274
    edge200                         = 1325
    fr910xt                         = 1328
    edge800_taiwan                  = 1333
    edge800_japan                   = 1334
    alf04                           = 1341
    fr610                           = 1345
    fr210_japan                     = 1360
    vector_ss                       = 1380
    vector_cp                       = 1381
    edge800_china                   = 1386
    edge500_china                   = 1387
    fr610_japan                     = 1410
    edge500_korea                   = 1422
    fr70                            = 1436
    fr310xt_4t                      = 1446
    amx                             = 1461
    fr10                            = 1482
    edge800_korea                   = 1497
    swim                            = 1499
    fr910xt_china                   = 1537
    Fenix                           = 1551
    edge200_taiwan                  = 1555
    edge510                         = 1561
    edge810                         = 1567
    Tempe                           = 1570
    fr910xt_japan                   = 1600
    GPS_1619                        = 1619
    GPS_1620                        = 1620
    GPS_1621                        = 1621
    fr620                           = 1623
    fr220                           = 1632
    fr910xt_korea                   = 1664
    fr10_japan                      = 1688
    edge810_japan                   = 1721
    virb_elite                      = 1735
    edge_touring                    = 1736
    edge510_japan                   = 1742
    HRM_Tri                         = 1743
    HRM_Run                         = 1752
    fr920xt                         = 1765
    edge510_asia                    = 1821
    edge810_china                   = 1822
    edge810_taiwan                  = 1823
    edge1000                        = 1836
    vivofit                         = 1837
    virb_remote                     = 1853
    vivo_ki                         = 1885
    fr15                            = 1903
    VivoActive                      = 1907
    edge510_korea                   = 1918
    fr620_japan                     = 1928
    fr620_china                     = 1929
    fr220_japan                     = 1930
    fr220_china                     = 1931
    approach_s6                     = 1936
    VivoSmart                       = 1956
    Fenix2                          = 1967
    epix                            = 1988
    Fenix3                          = 2050
    edge1000_taiwan                 = 2052
    edge1000_japan                  = 2053
    fr15_japan                      = 2061
    edge520                         = 2067
    edge1000_china                  = 2070
    fr620_russia                    = 2072
    fr220_russia                    = 2073
    vector_s                        = 2079
    edge1000_korea                  = 2100
    fr920xt_taiwan                  = 2130
    fr920xt_china                   = 2131
    fr920xt_japan                   = 2132
    virbx                           = 2134
    vivo_smart_apac                 = 2135
    etrex_touch                     = 2140
    edge25                          = 2147
    fr25                            = 2148
    VivoFit2                        = 2150
    fr225                           = 2153
    Forerunner_630                  = 2156
    fr230                           = 2157
    vivo_active_apac                = 2160
    vector_2                        = 2161
    vector_2s                       = 2162
    virbxe                          = 2172
    fr620_taiwan                    = 2173
    fr220_taiwan                    = 2174
    truswing                        = 2175
    Fenix3_china                    = 2188
    Fenix3_twn                      = 2189
    varia_headlight                 = 2192
    varia_taillight_old             = 2193
    edge_explore_1000               = 2204
    fr225_asia                      = 2219
    varia_radar_taillight           = 2225
    varia_radar_display             = 2226
    edge20                          = 2238
    D2_Bravo                        = 2262
    approach_s20                    = 2266
    varia_remote                    = 2276
    HRM4_Run                        = 2327
    VivoActive_HR                   = 2337
    VivoSmart_GPS_HR                = 2347
    VivoSmart_HR                    = 2348
    VivoMove                        = 2368
    varia_vision                    = 2398
    VivoFit3                        = 2406
    Fenix3_HR                       = 2413
    Index_Smart_Scale               = 2429
    fr235                           = 2431
    Fenix3_Chronos                  = 2432
    Oregon7xx                       = 2441
    Rino_7xx                        = 2444
    nautix                          = 2496
    Forerunner35                    = 2503
    Edge_820                        = 2530
    Edge_Explore_820                = 2531
    Fenix_5S                        = 2544
    D2_Bravo_Titanium               = 2547
    Varia_UT800                     = 2567
    Running_Dynamics_Pod            = 2593
    Fenix5X                         = 2604
    VivoFit_Jr                      = 2606
    VivoSport                       = 2623
    Forerunner935                   = 2691
    Fenix_5_Sapphire                = 2697
    VivoActive_3                    = 2700
    Edge_1030                       = 2713
    VivoMove_HR                     = 2772
    Approach_z80                    = 2806
    VivoSmart3_Apac                 = 2831
    VivoSport_Apac                  = 2832
    Descent                         = 2859
    Forerunner_645                  = 2886
    Forerunner_645m                 = 2888
    Fenix_5S_Plus                   = 2900
    Edge_130                        = 2909
    Edge_1030_Asia                  = 2924
    Vivosmart_4                     = 2927
    VivoMove_HR_Asia                = 2945
    GPS_2957                        = 2957
    Approach_x10                    = 2962
    ForeRunner_30_Asia              = 2977
    VivoActive_3m_w                 = 2988
    ForeRunner_645_Asia             = 3003
    ForeRunner_645M_Asia            = 3004
    Edge_Explore                    = 3011
    GPSMap66                        = 3028
    Approach_S10                    = 3049
    VivoActive_3M_l                 = 3066
    Approach_G80                    = 3085
    Edge_130_Asia                   = 3092
    Edge_1030_Bontrager             = 3095
    GPS_3107                        = 3107
    Fenix_5_Plus                    = 3110
    Fenix_5x_Plus                   = 3111
    Edge_520_Plus                   = 3112
    ForeRunner_945                  = 3113
    Edge_530                        = 3121
    Edge_830                        = 3122
    Instinct_Esports                = 3126
    Fenix_5S_Plus_Apac              = 3134
    Fenix_5X_Plus_Apac              = 3135
    Edge_520_Plus_Apac              = 3142
    Forerunner235L_Asia             = 3144
    Forerunner235_Asia              = 3145
    VivoActive_3m_Asia              = 3163
    Bike_Speed_Sensor_Gen3          = 3192
    Bike_Cadence_Sensor_Gen3        = 3193
    VivoSmart_4_Asia                = 3218
    VivoActive_4_Small              = 3224
    VivoActive_4_Large              = 3225
    Venu                            = 3226
    Marq_Driver                     = 3246
    Marq_Aviator                    = 3247
    Marq_Captain                    = 3248
    Marq_Commander                  = 3249
    Marq_Expedition                 = 3250
    Marq_Athlete                    = 3251
    Descent_Mk2                     = 3258
    Fenix_6S_Sport                  = 3287
    Fenix_6S_Pro                    = 3288
    Fenix_6_Sport                   = 3289
    Fenix_6                         = 3290
    Fenix_6X                        = 3291
    sensor_hub_3294                 = 3294
    HRM_Dual                        = 3299
    Approach_S40                    = 3314
    ForeRunner_245M_Asia            = 3321
    Edge_530_Apac                   = 3349
    Edge_830_Apac                   = 3350
    VivoMove_3                      = 3378
    VivoActive_4_Small_Asia         = 3387
    VivoActive_4_Large_Asia         = 3388
    VivoActive_4_OLED_Asia          = 3389
    SWIM2                           = 3405
    Marq_Driver_Asia                = 3420
    Marq_Aviator_Asia               = 3421
    VivoMove_3_Asia                 = 3422
    ForeRunner_945_Asia             = 3441
    VivoActive_3T_CHN               = 3446
    Marq_Captain_Asia               = 3448
    Marq_Commander_Asia             = 3449
    Marq_Expedition_Asia            = 3450
    Marq_Athlete_Asia               = 3451
    ForeRunner_45_Asia              = 3469
    VivoActive_3_Daimler            = 3473
    Fenix_6S_Sport_Asia             = 3512
    Fenix_6S_Asia                   = 3513
    Fenix_6_Sport_Asia              = 3514
    Fenix_6_Asia                    = 3515
    Fenix_6X_Asia                   = 3516
    Edge_130_Plus                   = 3558
    Edge_1030_Plus                  = 3570
    ForeRunner_745                  = 3589
    VenusQ                          = 3600
    Marq_Adventurer                 = 3624
    Marq_Adventurer_Asia            = 3648
    SWIM2_Apac                      = 3639
    Descent_Mk2_Asia                = 3702  # Mk2 and Mk2i
    Venu_Daimler_Asia               = 3737
    Marq_Golfer                     = 3739
    Venu_Daimler                    = 3740
    ForeRunner_745_Asia             = 3794
    Edge_1030_Plus_Asia             = 3812
    Edge_130_Plus_Asia              = 3813
    VenusQ_Asia                     = 3837
    Marq_Golfer_Asia                = 3850
    Venu_2_Plus                     = 3851
    FootPod_SDM4                    = 10007
    Edge_Remote                     = 10014
    TACX_Training_App_Win           = 20533
    TACX_Training_App_Mac           = 20534
    Training_Center                 = 20119
    TACX_Training_App_Android       = 30045
    TACX_Training_Aapp_IOS          = 30046
    TACX_Training_App_Legacy        = 30047
    Connectiq_Simulator             = 65531
    Android_Antplus_plugin          = 65532
    connect                         = 65534
    invalid                         = 65535


class GarminLocalProduct(FuzzyFieldEnum):
    """Garmin product codes used in FIT files for sub-devices."""

    Bluetooth_Low_Energy_Chipset    = 0
    Accelerometer_12                = 12
    Accelerometer_16                = 16
    Accelerometer_18                = 18
    Accelerometer_865               = 865
    Accelerometer_901               = 901
    GPS_1619                        = 1619
    GPS_1620                        = 1620
    GPS_1621                        = 1621
    GPS_2957                        = 2957
    Accelerometer_8194              = 8194
    Accelerometer_8191              = 8191
    Accelerometer_8195              = 8195
    Accelerometer_12529             = 12529
    Accelerometer_12533             = 12533
    Accelerometer_16150             = 16150
    Accelerometer_16155             = 16155
    Accelerometer_16310             = 16310
    Accelerometer_16423             = 16423
    Accelerometer_16427             = 16427
    Accelerometer_16354             = 16354
    Accelerometer_17350             = 17530
    Accelerometer_21909             = 21909
    BTLE_Chipset                    = 24832
    Accelerometer_49208             = 49208


class WahooFitnessProduct(FuzzyFieldEnum):
    """Wahoo Fitness product codes used in FIT files."""

    RPM_Sensor = 6


class ScoscheProduct(FuzzyFieldEnum):
    """Scosche product codes used in FIT files."""

    Rhythm_Plus_Armband_HRM = 2


class HealthAndLifeProduct(FuzzyFieldEnum):
    """Scosche product codes used in FIT files."""

    Accelerometer_515 = 515


class UnknownProduct(UnknownEnumValue):
    """Unknown product codes used in FIT files."""


def product_enum(manufacturer, product_str):
    """Return a product enum given the manufacturer enum and a product string."""
    _manufacturer_to_product_enum = {
        Manufacturer.Garmin                 : GarminProduct,
        Manufacturer.Garmin_local           : GarminLocalProduct,
        Manufacturer.Dynastream             : GarminProduct,
        Manufacturer.Dynastream_OEM         : GarminProduct,
        Manufacturer.Wahoo_Fitness          : WahooFitnessProduct,
        Manufacturer.Scosche                : ScoscheProduct,
        Manufacturer.Health_and_Life        : HealthAndLifeProduct,
        Manufacturer.invalid                : GarminProduct,
    }
    return _manufacturer_to_product_enum[manufacturer].from_string(product_str)



"""Methods for converting metrics from one representation to another."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


import datetime
import string


def ms_to_dt_time(time_ms):
    """Convert time in milli seconds to a datetime object."""
    if time_ms is not None:
        return (datetime.datetime.min + datetime.timedelta(milliseconds=time_ms)).time()


def secs_to_dt_time(time_secs):
    """Convert time in seconds to a datetime object."""
    if time_secs is not None:
        return (datetime.datetime.min + datetime.timedelta(seconds=time_secs)).time()


def min_to_dt_time(time_mins):
    """Convert time in minutes to a datetime object."""
    if time_mins is not None:
        return (datetime.datetime.min + datetime.timedelta(minutes=time_mins)).time()


def hours_to_dt_time(time_hours):
    """Convert time in hours to a datetime object."""
    if time_hours is not None:
        return (datetime.datetime.min + datetime.timedelta(hours=time_hours)).time()


def time_to_secs(time):
    """Convert a datetime time object to seconds."""
    if time is not None:
        return (((time.hour * 60) + time.minute) * 60) + time.second


def time_to_timedelta(time):
    """Convert a datetime time object to timedelta object."""
    if time is not None:
        return datetime.timedelta(hours=time.hour, minutes=time.minute, seconds=time.second)


def timedelta_to_time(timedelta):
    """Convert a timedelta object to datetime time object."""
    return (datetime.datetime.min + timedelta).time()


def add_time(time1, time2, scale=1):
    """Add two datetime time values, treating both as elapsed time."""
    return (datetime.datetime.min + time_to_timedelta(time1) + (time_to_timedelta(time2) * scale)).time()


def subtract_time(time1, time2):
    """Subtract two datetime time values, treating both as elapsed time."""
    return (datetime.datetime.min + time_to_timedelta(time1) - time_to_timedelta(time2)).time()


def day_of_the_year_to_datetime(year, day):
    """Given the index of a day within the year, return a datetime object."""
    return datetime.datetime(year, 1, 1) + datetime.timedelta(day - 1)


def meters_to_feet(meters):
    """Convert meters to feet."""
    if meters is not None:
        return (meters * 3.2808399)


def meters_to_miles(meters):
    """Convert meters to miles."""
    if meters is not None:
        return meters_to_feet(meters) / 5280.0


def mps_to_mph(meters_per_sec):
    """Convert meters per second to miles per hour."""
    if meters_per_sec is not None:
        return meters_per_sec * 2.236936


def celsius_to_fahrenheit(celsius):
    """Convert celsius_to_fahrenheit to fahrenheit."""
    if celsius is not None:
        return (celsius * 1.8) + 32.0


def date_to_dt(date):
    """Given a datetime date object, return a date time datetime object for the given date at 00:00:00."""
    return datetime.datetime.combine(date, datetime.time.min)


def dt_to_epoch_ms(dt):
    """Convert a datetime object to milliseconds since the epoch."""
    return int((dt - datetime.datetime.fromtimestamp(0)).total_seconds() * 1000)


def dt_to_utc_epoch_ms(dt):
    """Convert a UTC datetime object to milliseconds since the epoch."""
    return int((dt - datetime.datetime.utcfromtimestamp(0)).total_seconds() * 1000)


def printable(string_in):
    """Return a string with only prinatable characters given a string with potentially unprintable characters."""
    if string_in is not None:
        return filter(lambda x: x in string.printable, string_in)


def persec_speed_to_pace(speed):
    """Convert a per second speed to a per distance pace."""
    if speed:
        # secs_per_hour / speed_per_hour = speed_per
        return (datetime.datetime.min + datetime.timedelta(seconds=(3600 / speed))).time()


def perhour_speed_to_pace(speed):
    """Convert a per second speed to a per distance pace."""
    if speed:
        # secs_per_hour / speed_per_hour = speed_per
        return (datetime.datetime.min + datetime.timedelta(seconds=(3600 / speed))).time()



"""Manages FIT file format base types."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


class BaseType():
    """Manages FIT file format base types."""

    __base_type_data = {
        0x00 : [False, 'enum',     0xFF,               'UINT8'],
        0x01 : [False, 'sint8',    0x7F,               'INT8'],
        0x02 : [False, 'uint8',    0xFF,               'UINT8'],
        0x07 : [False, 'string',   0x00,               'CHAR'],
        0x83 : [True,  'sint16',   0x7FFF,             'INT16'],
        0x84 : [True,  'uint16',   0xFFFF,             'UINT16'],
        0x85 : [True,  'sint32',   0x7FFFFFFF,         'INT32'],
        0x86 : [True,  'uint32',   0xFFFFFFFF,         'UINT32'],
        0x88 : [True,  'float32',  0xFFFFFFFF,         'FLOAT32'],
        0x89 : [True,  'float64',  0xFFFFFFFFFFFFFFFF, 'FLOAT64'],
        0x0a : [False, 'uint8z',   0x00,               'UINT8'],
        0x8b : [True,  'uint16z',  0x00000000,         'UINT16'],
        0x8c : [True,  'uint32z',  0x00000000,         'UINT32'],
        0x0d : [False, 'byte',     0xFF,               'UINT8'],
        0x8e : [True,  'sint64',   0x7FFFFFFFFFFFFFFF, 'INT64'],
        0x8f : [True,  'uint64',   0xFFFFFFFFFFFFFFFF, 'UINT64'],
        0x90 : [True,  'uint64z',  0x0000000000000000, 'UINT64'],
    }
    index_endian = 0
    index_name = 1
    index_invalid = 2
    index_string = 3

    @classmethod
    def __base_type(cls, index):
        try:
            return cls.__base_type_data[index]
        except Exception:
            raise IndexError(f'Unknown base type index {index}')

    @classmethod
    def _type_endian(cls, index):
        return cls.__base_type(index)[cls.index_endian]

    @classmethod
    def type_name(cls, index):
        """Return the name for a type given it's index."""
        return cls.__base_type(index)[cls.index_name]

    @classmethod
    def _invalid(cls, index):
        """Return the invalid value for a type given it's index."""
        return cls.__base_type(index)[cls.index_invalid]

    @classmethod
    def _type_string(cls, index):
        """Return the name of a type given it's index."""
        return cls.__base_type(index)[cls.index_string]



"""Test FIT file parsing."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import unittest
import logging
import datetime

from fitfile import fields


root_logger = logging.getLogger()
handler = logging.FileHandler('fit_fields.log', 'w')
root_logger.addHandler(handler)
root_logger.setLevel(logging.INFO)

logger = logging.getLogger(__name__)


class TestFitFields(unittest.TestCase):
    """Class for testing FIT file parsing."""

    @classmethod
    def setUpClass(cls):
        pass

    def test_time_ms_field_valid_conversion(self):
        time = fields.TimeMsField('time')
        field_value_list = time.convert(508324, 4294967295)
        self.assertEqual(field_value_list[0]['time'], datetime.time(0, 8, 28, 324000))

    def __utc_timestamp_field_valid_conversion(self, value, expected_dt):
        time = fields.TimestampField('timestamp', utc=True)
        field_value_list = time.convert(value, 4294967295)
        self.assertEqual(field_value_list[0]['timestamp'], expected_dt)

    def test_utc_timestamp_field_valid_conversion(self):
        test_values = [
            (936189712, datetime.datetime(2019, 8, 31, 12, 41, 52, tzinfo=datetime.timezone.utc)),
            (936188666, datetime.datetime(2019, 8, 31, 12, 24, 26, tzinfo=datetime.timezone.utc)),
            (943190700, datetime.datetime(2019, 11, 20, 13, 25, 0, tzinfo=datetime.timezone.utc)),
            (857947331, datetime.datetime(2017, 3, 8, 22, 42, 11, tzinfo=datetime.timezone.utc))
        ]
        for value, expected_dt in test_values:
            self.__utc_timestamp_field_valid_conversion(value, expected_dt)

    def test_local_timestamp_field_valid_conversion(self):
        time = fields.TimestampField('timestamp', utc=False)
        field_value_list = time.convert(936189712, 4294967295)
        self.assertEqual(field_value_list[0]['timestamp'], datetime.datetime(2019, 8, 31, 12, 41, 52))


if __name__ == '__main__':
    unittest.main(verbosity=2)



"""Objects that represent exceptions while parsing a FIT file."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


class FitException(Exception):
    """Base class for FIT file exceptions."""

    def __init__(self, message):
        """Return a FitException instance."""
        self.message = message

    def __str__(self):
        """Return a string representation of a FitException instance."""
        return self.message


class FitFileError(FitException):
    """An exception when parsing a FIT file."""


class FitFileBadHeaderSize(FitFileError):
    """An exception do to a bad FIT file header size."""


class FitFileBadProtocolVersion(FitFileError):
    """An exception do to a bad FIT file protocol version."""


class FitFileDataType(FitFileError):
    """An exception happenbed will parsing FIT file data."""


class FitMessageType(FitFileError):
    """An exception do to a unrecognized FIT file message type."""


class FitMessageParse(FitFileError):
    """An exception happenbed will parsing a FIT file message."""


class FitUndefDevMessageType(FitFileError):
    """An undefined message type was encountered."""


class FitOutOfOrderMessage(FitFileError):
    """The message's timestamp preceeded the previous message's."""



"""Test FIT file parsing."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import unittest
import logging

from fitfile import field_enums
from fitfile import enum_fields


root_logger = logging.getLogger()
handler = logging.FileHandler('fit_field_enum.log', 'w')
root_logger.addHandler(handler)
root_logger.setLevel(logging.INFO)

logger = logging.getLogger(__name__)


class TestFitFieldEnum(unittest.TestCase):
    """Class for testing FIT file parsing."""

    @classmethod
    def setUpClass(cls):
        pass

    def test_field_enum_valid_conversion(self):
        self.assertEqual(field_enums.Switch.from_string('on'), field_enums.Switch.on)

    def test_field_enum_unknown_conversion(self):
        self.assertIsInstance(field_enums.Switch.from_string('junk'), field_enums.UnknownEnumValue)

    def test_field_enum_fuzzy_metric(self):
        self.assertEqual(field_enums.DisplayMeasure.from_string('metric_system'), field_enums.DisplayMeasure.metric)

    def test_field_enum_fuzzy_statute(self):
        self.assertEqual(field_enums.DisplayMeasure.from_string('statute_us'), field_enums.DisplayMeasure.statute)

    def test_enum_field_valid_conversion(self):
        switch = enum_fields.SwitchField('test')
        field_value_list = switch.convert(1, 255)
        self.assertEqual(field_value_list[0]['test'], field_enums.Switch.on)

    def test_enum_field_unknown_conversion(self):
        switch = enum_fields.SwitchField('test')
        field_value_list = switch.convert(10, 255)
        self.assertIsInstance(field_value_list[0]['test'], field_enums.UnknownEnumValue)


if __name__ == '__main__':
    unittest.main(verbosity=2)



"""FIT file definition message."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


import collections

from .data import Schema, Data, Architecture
from .fields import UnknownField
from .developer_field_definition import DeveloperFieldDefinition
from .definition_message_data import DefinitionMessageData
from .field_definition import FieldDefinition
from .message_type import MessageType


class DefinitionMessage(Data):
    """FIT file definition message."""

    dm_primary_schema = Schema(
        'dm_primary',
        collections.OrderedDict(
            [
                ('reserved', ['UINT8', 1]),
                ('architecture', ['UINT8', 1])
            ]
        )
    )
    dm_secondary_schema = Schema(
        'dm_secondary',
        collections.OrderedDict(
            [
                ('global_message_number', ['UINT16', 1]),
                ('fields', ['UINT8', 1])
            ]
        )
    )
    dm_dev_schema = Schema(
        'dm_dev',
        collections.OrderedDict(
            [
                ('dev_fields', ['UINT8', 1])
            ]
        )
    )

    def __init__(self, record_header, dev_field_dict, file):
        """
        Return a DefinitionMessage instance created by reading data from a FIT file.

        Paramters:
            record_header (RecordHeader): the record header associated with this definition message.
            dev_field_dict (dict): a dictionary of developer defoined fields in the FIT file.
            file (File): the FIT file instance to read the definition message data from.
        """
        self.reserved = None
        self.architecture = None
        self.global_message_number = None
        self.fields = None
        self.dev_fields = None
        super().__init__(file, DefinitionMessage.dm_primary_schema, [(DefinitionMessage.dm_secondary_schema, self.__decode_secondary)])

        self.message_type = MessageType.get_type(self.global_message_number)
        self.__message_data = DefinitionMessageData.get_message_definition(self.message_type)

        self.field_definitions = []
        for _ in range(self.fields):
            field_definition = FieldDefinition(file)
            self.file_size += field_definition.file_size
            self.field_definitions.append(field_definition)

        self.has_dev_fields = record_header.developer_data()
        self.dev_field_definitions = []
        if self.has_dev_fields:
            self._decode(DefinitionMessage.dm_dev_schema)
            for _ in range(self.dev_fields):
                dev_field_definition = DeveloperFieldDefinition(dev_field_dict, file)
                self.file_size += dev_field_definition.file_size
                self.dev_field_definitions.append(dev_field_definition)

    def __decode_secondary(self):
        try:
            self.endian = Architecture(self.architecture)
        except Exception:
            # treat all broken endians as little endian
            self.endian = Architecture.Little_Endian
        return True

    def field(self, field_number):
        """Return an instance of the proper Field subclass for the given field definition."""
        return DefinitionMessageData.reserved_field_indexes.get(field_number, self.__message_data.get(field_number, UnknownField(field_number)))

    def __str__(self):
        """Return a string representation of a DefinitionMessage instance."""
        return (f"DefinitionMessage: {repr(self.message_type)} {self.fields} {self.endian.name} fields")



"""Library for parsing FIT files and returning them as lists of messages dictionaries."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

# flake8: noqa

from .conversions import ms_to_dt_time, secs_to_dt_time, min_to_dt_time, hours_to_dt_time, time_to_secs, time_to_timedelta, timedelta_to_time, add_time, subtract_time, day_of_the_year_to_datetime, meters_to_feet, meters_to_miles, mps_to_mph, celsius_to_fahrenheit, date_to_dt, dt_to_epoch_ms, dt_to_utc_epoch_ms, printable, perhour_speed_to_pace



"""Test conversion functions."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import unittest
import logging
import datetime

from fitfile import conversions


root_logger = logging.getLogger()
handler = logging.FileHandler('conversions.log', 'w')
root_logger.addHandler(handler)
root_logger.setLevel(logging.INFO)

logger = logging.getLogger(__name__)


class TestConversions(unittest.TestCase):
    """Class for testing conversion functions."""

    @classmethod
    def setUpClass(cls):
        pass

    def test_perhour_speed_to_pace(self):
        test_data = [
            (3.0, datetime.time(minute=20)),
            (5.0, datetime.time(minute=12)),
            (7.2, datetime.time(minute=8, second=20))
        ]
        for (speed, pace) in test_data:
            converted_pace = conversions.perhour_speed_to_pace(speed)
            self.assertEqual(converted_pace, pace, f'Bad conversion for speed {speed}')


if __name__ == '__main__':
    unittest.main(verbosity=2)



"""Objects that represent FIT file activity message fields."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


from .fields import Field, NamedField, CyclesField, StepsField, StrokesField
from .enum_fields import EnumField
from .field_enums import Activity, ActivityType
from .sport import Sport, SubSport


def _cycles_name_to_field(activity):
    field_mapping = {
        'generic'                   : CyclesField,
        # steps activities
        'walking'                   : StepsField,
        'running'                   : StepsField,
        'hiking'                    : StepsField,
        'elliptical'                : StepsField,
        # strokes activities
        'cycling'                   : StrokesField,
        'swimming'                  : StrokesField,
        'rowing'                    : StrokesField,
        'paddling'                  : StrokesField,
        'stand_up_paddleboarding'   : StrokesField,
        'kayaking'                  : StrokesField,
    }
    try:
        return field_mapping[activity.name]
    except Exception:
        return CyclesField


class ActivityBasedCyclesField(NamedField):
    """A cycles field that gnerates dependant fields based on the activity type."""

    _name = 'cycles'
    _units = 'cycles'
    _scale = 2.0
    _dependant_field_control_fields = ['activity_type']

    def dependant_field(self, control_value_list):
        """Return a dependant field instance given the control field values."""
        activity_type = control_value_list[0]
        dependant_field = _cycles_name_to_field(activity_type)
        dependant_field_name = self.name.replace('cycles', dependant_field._units)
        return dependant_field(name=dependant_field_name)


class ActivityField(EnumField):
    """A field holding an activity as a integer enum value."""

    _name = 'activity'
    _enum = Activity


class ActivityTypeField(Field):
    """A field holding an activity type as a integer enum value."""

    _name = 'activity_type'

    def _convert_single(self, value, invalid):
        return ActivityType(value)


class ActivityClassField(Field):
    """A field holding an activity class as a integer enum value."""

    _name = 'activity_class'

    def _convert_single(self, value, invalid):
        if value & 0x80:
            activity_class = "athlete "
        else:
            activity_class = ""
        activity_class += str(value & 0x7f)
        return activity_class


class IntensityField(Field):
    """A field that indicates how active the user was."""

    _name = 'intensity'
    _max_intensity = 8


class ActivityTypeIntensityField(NamedField):
    """A field that generates sub fields fields for activity and intensity."""

    activity_type_field = ActivityTypeField()
    intensity_field = IntensityField()

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)

    def convert(self, value, invalid, measurement_system):
        """Convert the value to sub fields."""
        self.measurement_system = measurement_system
        activity_type = value & 0x1f
        intensity = value >> 5
        return self.activity_type_field.convert(activity_type, 0xff, measurement_system) + self.intensity_field.convert(intensity, 0xff, measurement_system)


class SportBasedCyclesField(NamedField):
    """A cycles field that generates dependant fields based on the sport type."""

    _units = 'cycles'
    _dependant_field_control_fields = ['sport', 'sub_sport']
    _scale_map = {
        CyclesField: 1.0,
        StepsField: 0.5,
        StrokesField: 1.0
    }

    def dependant_field(self, control_value_list):
        """Return a dependant field instance given the control field values."""
        sport = control_value_list[0]
        dependant_field = _cycles_name_to_field(sport)
        if dependant_field._units == 'cycles':
            sub_sport = control_value_list[1]
            dependant_field = _cycles_name_to_field(sub_sport)
        dependant_field_name = self.name.replace('cycles', dependant_field._units)
        return dependant_field(name=dependant_field_name, scale=self._scale_map[dependant_field])


class CadenceField(NamedField):
    """Field that holds cycles/minute measurement for sports activity."""

    _name = 'cadence'
    _units = 'rpm'


class StepsCadenceField(NamedField):
    """Field that holds cycles/minute measurement for sports activity."""

    _name = 'steps_per_min'
    _units = 'steps/min'


class StrokesCadenceField(NamedField):
    """Field that holds cycles/minute measurement for sports activity."""

    _name = 'strokes_per_min'
    _units = 'strokes/min'


def _cadence_sport_to_field(activity):
    _units = {
        'generic'                   : CadenceField,
        # steps activities
        'walking'                   : StepsCadenceField,
        'running'                   : StepsCadenceField,
        'hiking'                    : StepsCadenceField,
        'elliptical'                : StepsCadenceField,
        # strokes activities
        'cycling'                   : StrokesCadenceField,
        'swimming'                  : StrokesCadenceField,
        'rowing'                    : StrokesCadenceField,
        'paddling'                  : StrokesCadenceField,
        'stand_up_paddleboarding'   : StrokesCadenceField,
        'kayaking'                  : StrokesCadenceField,
    }
    try:
        return _units[activity.name]
    except Exception:
        return _units['generic']


class SportBasedCadenceField(CadenceField):
    """A cycles field that generates dependant fields based on the sport type."""

    _dependant_field_control_fields = ['sport', 'sub_sport']
    _scale_map = {
        CadenceField: 1.0,
        StepsCadenceField: 0.5,
        StrokesCadenceField: 1.0
    }

    def dependant_field(self, control_value_list):
        """Return a dependant field instance given the control field values."""
        sport = control_value_list[0]
        dependant_field = _cadence_sport_to_field(sport)
        if dependant_field._units == 'rpm':
            sub_sport = control_value_list[1]
            dependant_field = _cadence_sport_to_field(sub_sport)
        dependant_field_name = self.name.replace('cadence', dependant_field._name)
        return dependant_field(name=dependant_field_name, scale=self._scale_map[dependant_field])


class SportField(EnumField):
    """A field representing a sport via an ineger enum value."""

    _name = 'sport'
    _enum = Sport
    _units = {
        0 : 'cycles',
        1 : 'steps',
        2 : 'strokes',
        5 : 'strokes',
        11 : 'steps',
        15 : 'strokes',
        17 : 'steps',
        19 : 'strokes',
        37 : 'strokes',
    }

    @classmethod
    def units(cls, sport_index):
        """Return the proper units for an activity given the sport index."""
        try:
            return cls._units[sport_index]
        except Exception:
            return cls._units[0]


class SubSportField(EnumField):
    """A field representing a sub-sport via an ineger enum value."""

    _name = 'sub_sport'
    _enum = SubSport



"""Objects defining FIT file developer fields."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


import collections
import logging

from .data import Schema
from .field_definition import FieldDefinitionBase
from .message_type import MessageType
from .definition_message_data import DefinitionMessageData
from .object_fields import DistanceMetersField, SpeedMpsField
from .dev_field import DevField, DevDistanceField, DerivedDevDistanceField, DevSpeedField, DerivedDevSpeedField
from .exceptions import FitUndefDevMessageType


logger = logging.getLogger(__name__)


class DeveloperFieldDefinition(FieldDefinitionBase):
    """Developer filed definitions decoded from a FIT file."""

    dfd_schema = Schema(
        'DeveloperFieldDefinition',
        collections.OrderedDict(
            [
                ('field_number', ['UINT8', 1]),
                ('size', ['UINT8', 1]),
                ('developer_data_index', ['UINT8', 1])
            ]
        )
    )

    def __init__(self, dev_field_dict, file):
        """
        Return a DeveloperFieldDefinition instance created by reading data from a FIT file.

        Paramters:
            dev_field_dict (dict): a dictionary of developer defined fields.
            file (File):  a FIT File instance.
        """
        self.field_number = None
        self.size = None
        self.developer_data_index = None
        super().__init__(file, DeveloperFieldDefinition.dfd_schema)
        self.dev_field_message = dev_field_dict.get(self.field_number)
        if self.dev_field_message is None:
            raise FitUndefDevMessageType(f'Dev field {self.field_number} undefined in {dev_field_dict}')
        # Parse values from dev_field_message
        self.field_name = self.dev_field_message.fields.field_name
        self.native_message_type = MessageType(self.dev_field_message.fields.native_message_num)
        self.native_field_num = self.dev_field_message.fields.native_field_num
        self.units = self.dev_field_message.fields.units
        self.offset = self.dev_field_message.fields.offset
        self.scale = self.dev_field_message.fields.scale
        self.base_type = self.dev_field_message.field_values.fit_base_type_id.orig
        # If the dev field shadows a native field, then take the field name from the native field.
        if self.native_message_type and self.native_field_num:
            field_dict = DefinitionMessageData.get_message_definition(self.native_message_type)
            field = field_dict[self.native_field_num]
            self.display_field_name = 'dev_' + field.name
            self._field = self.__derive_field(self.display_field_name, self.units, self.scale, self.offset, field)
        else:
            self.display_field_name = 'dev_' + self.field_name
            self._field = self.__map_field(self.display_field_name, self.units, self.scale, self.offset)
        logger.info('%s for %r field %s', self, self.native_message_type, self.native_field_num)

    @classmethod
    def __derive_field(cls, field_name, units, scale, offset, field_obj):
        if isinstance(field_obj, DistanceMetersField):
            return DerivedDevDistanceField(field_name, units, scale, offset, field_obj)
        if isinstance(field_obj, SpeedMpsField):
            return DerivedDevSpeedField(field_name, units, scale, offset, field_obj)
        return DevField(field_name, units, scale, offset)

    @classmethod
    def __map_field(cls, field_name, units, scale, offset):
        field_map = {
            'dev_distance'  : DevDistanceField,
            'dev_speed'     : DevSpeedField
        }
        if field_name in field_map:
            return field_map[field_name](field_name, units, scale, offset)
        return DevField(field_name, units, scale, offset)

    def field(self):
        """Return a field instance representing the field for this DeveloperFieldDefinition instance."""
        return self._field

    def __str__(self):
        """Return a string representation for the DeveloperFieldDefinition instance."""
        return (f'{self.__class__.__name__}({self.display_field_name}[{self.field_name}]: '
                + f'scale {self.scale} offset {self.offset} units {self.units}, {self.type_count()} of {self.type_string()})')



"""Objects that represent FIT file message fields."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


import datetime

from .conversions.conversions import ms_to_dt_time, hours_to_dt_time, min_to_dt_time, secs_to_dt_time
from .field_enums import DisplayMeasure
from .field_value import FieldValue


class Field(object):
    """The base object for all FIT file message fields."""

    _name = None
    _units = None
    _scale = 1.0
    _offset = 0.0
    _dependant_field_control_fields = None

    def __init__(self, **kwargs):
        """Return a new instance of the Field class."""
        for key, value in kwargs.items():
            vars(self)['_' + key] = value
        if self._name is None:
            raise ValueError(f'Unamed instance of {self.__class__.__name__}')

    @property
    def name(self):
        """Return the name of the field."""
        return self._name

    @property
    def units(self):
        """Return the units of the field."""
        return self._units

    def _invalid_single(self, value, invalid):
        return (value == invalid)

    def _invalid_many(self, values, invalid):
        for value in values:
            if self._invalid_single(value, invalid):
                return True
        return False

    def is_invalid(self, value, invalid):
        """Return if the field's value is valid."""
        if isinstance(value, list):
            return self._invalid_many(value, invalid)
        return self._invalid_single(value, invalid)

    def _convert_single(self, value, invalid):
        if value is not None and value != invalid:
            return (value / self._scale) + self._offset

    def __convert_many(self, _convert_single, value, invalid):
        if isinstance(value, list):
            return [_convert_single(sub_value, invalid) for sub_value in value]
        return _convert_single(value, invalid)

    def _convert_many(self, value, invalid):
        return self.__convert_many(self._convert_single, value, invalid)

    def convert(self, value, invalid, measurement_system=DisplayMeasure.metric):
        """Return a FieldValue as intepretted by the field's rules."""
        return [FieldValue(self, value, invalid, **{self._name: self._convert_many(value, invalid)})]

    def reconvert(self, value, invalid, measurement_system=DisplayMeasure.metric):
        """Return the field's value as intepretted by the field's rules."""
        return {self._name: self._convert_many(value, invalid)}

    def __repr__(self):
        """Return a string representation of a Field instance."""
        return f'{self.__class__.__name__} ({self._name})'


class NamedField(Field):
    """A field with a name that is passed to the constructor."""

    def __init__(self, *args, **kwargs):
        """Return an NamedField instance."""
        if len(args) > 0:
            super().__init__(name=args[0], **kwargs)
        else:
            super().__init__(**kwargs)


#
# Special fields
#
class UnknownField(Field):
    """Class that handles fields that are not documented."""

    def __init__(self, index):
        """Return a new instance of the UnknownField class."""
        super().__init__(name=f"unknown_{index}")


#
# Basic field types
#
class LeftRightBalanceField(NamedField):
    """A composite field that indicates left or roight and the percentage for that side."""

    def _convert_single(self, value, invalid):
        if value != invalid:
            if value & 0x8000:
                left_or_right = 'Right'
            else:
                left_or_right = 'Left'
            percentage = (value & 0x3fff) / 100
            return f'{left_or_right} {percentage} %'


class PercentField(NamedField):
    """A field holding a integer percentage value."""

    _units = '%'
    _scale = 100.0


class BytePercentField(PercentField):
    """A field holding a integer percentage value."""

    _scale = 2.0


class MessageIndexField(NamedField):

    def _convert_single(self, value, invalid):
        return {'selected': ((value & 0x8000) == 0x8000), 'value': (value & 0x0FFF)}


class CaloriesField(NamedField):
    """A field containing a calories measurement in kcal."""

    _name = 'calories'
    _units = 'kcal'


class CaloriesDayField(NamedField):
    """A field containing a calories measurement for a day in kcal/day."""

    _units = 'kcal/day'


class CyclesCaloriesField(Field):

    _name = 'cycles_to_calories'
    _units = 'kcal/cycle'
    _scale = 5019.6


class CyclesDistanceField(Field):

    _name = 'cycles_to_distance'
    _units = 'm/cycle'
    _scale = 5000.0


#
# Time related fields
#
class TimestampField(NamedField):

    _utc = None

    def _convert_single(self, value, invalid):
        if self._utc:
            return datetime.datetime(1989, 12, 31, 0, 0, 0, tzinfo=datetime.timezone.utc) + datetime.timedelta(seconds=value)
        return datetime.datetime(1989, 12, 31, 0, 0, 0) + datetime.timedelta(seconds=value)


class TimeMsField(NamedField):
    """A field holsing milliseconds returned as a datetime."""

    def _convert_single(self, value, invalid):
        if value != invalid:
            return ms_to_dt_time(value / self._scale)


class TimeSField(NamedField):
    """A field holding an integer number of seconds."""

    _units = 's'

    # invalid is not allowed, 65535 is a valid value
    def _convert_single(self, value, invalid):
        return value


class TimeOffsetField(Field):

    _name = 'time_offset'
    _units = 's'

    def _convert_single(self, value, invalid):
        # if the offset is greater than 24 hours, than it's negative
        if value > 86400:
            return -(0xFFFFFFFF - value + 1)
        return value


class TimeHourField(TimeMsField):
    """A field that holds a time value measured in milliseconds."""

    def _convert_single(self, milliseconds, invalid):
        if milliseconds != invalid:
            return hours_to_dt_time(milliseconds / self._scale)


class TimeMinField(TimeMsField):
    """A field that holds a time value measured in minutes."""

    def _convert_single(self, minutes, invalid):
        if minutes != invalid:
            return min_to_dt_time(minutes / self._scale)


class TimeOfDayField(NamedField):
    """A field that holds a time value measured in seconds."""

    def _convert_single(self, seconds, invalid):
        if seconds != invalid:
            return secs_to_dt_time(seconds)


class CyclesField(Field):
    """Field that holds cycles measurement for sports activity."""

    _name = 'cycles'
    _units = 'cycles'
    _scale = 2.0


class StepsField(Field):
    """Field that holds steps measurement for sports activity."""

    _name = 'steps'
    _units = 'steps'


class StrokesField(Field):
    """Field that holds strokes measurement for sports activity."""

    _name = 'strokes'
    _units = 'strokes'
    _scale = 2.0


class FractionalCyclesField(Field):
    """Field that holds cycles measurement for sports activity."""

    _name = 'total_fractional_cycles'
    _units = 'cycles'
    _scale = 128.0


class VersionField(NamedField):
    """A field that contains a software or hardware version."""

    _name = 'version'
    _scale = 100.0

    def _convert_single(self, value, invalid):
        if value != invalid:
            return '{0:2.2f}'.format(value / self._scale)


class EventDataField(Field):
    """A field that holds data that depends on the event's type."""

    _name = 'event_data'
    _dependant_field = {
        'timer' : Field(name='timer_trigger'),
        245     : CyclesField
    }
    _dependant_field_control_fields = ['event']

    def dependant_field(self, control_value_list):
        """Return a field whose type is based on the event type."""
        event = control_value_list[0]
        return EventDataField._dependant_field[event]


class FractionalCadenceField(NamedField):

    _name = 'fractional_cadence'
    _units = 'rpm'
    _scale = 128.0


class PowerField(NamedField):
    """A field that holds a power measurement."""

    _name = 'power'
    _units = 'watts'


class WorkField(Field):
    """A field that holds a work measurement in joules for an activity or portion of an activity."""

    _name = 'total_work'
    _units = 'J'


class TrainingEffectField(NamedField):
    """A field that holds a Garmin training effect measurement (0.0-5.0)."""

    _scale = 10.0



"""Objects that represent a FIT file record header."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


import collections
import enum

from .data import Schema, Data


class MessageClass(enum.Enum):
    """Enum reprersenting the class of a message."""

    data        = 0
    definition  = 1


class RecordHeader(Data):
    """Object that represents a FIT file record header."""

    rh_schema = Schema(
        'RecordHeader',
        collections.OrderedDict([('record_header', ['UINT8', 1])])
    )
    message_type_string = ['data', 'definition']

    def __init__(self, file):
        """Return a RecordHeader instance created by reading the record header data from a FIT file."""
        self.record_header = None
        super().__init__(file, self.rh_schema)
        self.message_class = MessageClass(self.message_type())

    def __compressed_timestamp(self):
        return (self.record_header & 0x80) == 0x80

    def message_type(self):
        """Return the type of the message."""
        return (self.record_header & 0x40) == 0x40

    def developer_data(self):
        """Return if the message contains developer data."""
        return (self.record_header & 0x60) == 0x60

    def local_message(self):
        """Return if the message is a local message."""
        return (self.record_header & 0x0f)

    def __str__(self):
        """Return a string representation of a RecordHeader instance."""
        return f'RecordHeader: Local {self.message_class.name} message {self.local_message()} (Compressed {self.__compressed_timestamp()})'

    def __repr__(self):
        """Return a string representation of a RecordHeader instance."""
        return self.__str__()



"""Class that represents a FIT file header."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


import collections

from .exceptions.exceptions import FitFileBadHeaderSize, FitFileBadProtocolVersion, FitFileDataType
from .data import Schema, Data


class FileHeader(Data):
    """Class that decodes a FIT file header."""

    fh_primary_schema = Schema(
        'fh_primary',
        collections.OrderedDict(
            [
                ('header_size', ['UINT8', 1]),
                ('protocol_version', ['UINT8', 1]),
                ('profile_version', ['UINT16', 1]),
                ('data_size', ['UINT32', 1]),
                ('data_type', ['CHAR', 4])
            ]
        )
    )
    fh_optional_schema = Schema(
        'fh_optional',
        collections.OrderedDict([('crc', ['UINT16', 1])])
    )
    profile_version_str = {100 : 'activity', 1602 : 'device'}

    min_file_header_size = 12
    opt_file_header_size = 14
    min_protocol_version = 0x10
    file_data_type = [46, 70, 73, 84]
#    file_data_type = ['.', 'F', 'I', 'T']

    def __init__(self, file):
        """Return a FileHeader instance created by reading data from a Fit file."""
        self.header_size = None
        self.protocol_version = None
        self.profile_version = None
        self.data_size = None
        self.data_type = None
        super().__init__(file, FileHeader.fh_primary_schema, [(FileHeader.fh_optional_schema, self.__decode_secondary)])
        self.__check()

    def __decode_secondary(self):
        return (self.header_size >= FileHeader.opt_file_header_size)

    def __check(self):
        if self.header_size < FileHeader.min_file_header_size:
            raise FitFileBadHeaderSize(f'{self.header_size} < {FileHeader.min_file_header_size}')
        if self.protocol_version < FileHeader.min_protocol_version:
            raise FitFileBadProtocolVersion(f'{self.protocol_version} < {FileHeader.min_protocol_version}')
        if self.data_type != FileHeader.file_data_type:
            raise FitFileDataType(f'{repr(self.data_type)} < {repr(FileHeader.file_data_type)}')

    def __str__(self):
        """Return a string representation of a FileHeader instance."""
        return f'{self.__class__.__name__}(header size {self.header_size} prot ver {self.protocol_version} prof ver {self.profile_version}'



"""Enums that represent FIT file message field values."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import enum


def name_for_enum(enum_instance):
    """Return the name if variable is enum or UnknownEnumValue."""
    return getattr(enum_instance, 'name', enum_instance)


class UnknownEnumValue(object):
    """Returned when a value can not be cast to an FieldEnum."""

    def __init__(self, value):
        """Return a UnknownEnumValue instance."""
        self.value = value
        self.name = f'{type(self).__name__}_{value}'

    @classmethod
    def from_string(cls, string):
        """Return a UnknownEnumValue instance created from a string."""
        return cls(string)

    def __eq__(self, other):
        """Test two UnknownEnumValue instances for equivelence."""
        return (other and (isinstance(other, UnknownEnumValue) and self.value == other.value)
                or (not isinstance(other, UnknownEnumValue) and self.value == other))

    def __ne__(self, other):
        """Test two UnknownEnumValue instances for equivelence."""
        return not self.__eq__(other)

    def __hash__(self):
        """Return a hash value for a UnknownEnumValue instance."""
        return self.value

    def __repr__(self):
        """Return a string representation of a UnknownEnumValue instance."""
        return f'<{type(self).__name__}.{self.name}: {self.value}>'


class FieldEnum(enum.Enum):
    """A enum representing a FIT file message field value."""

    @classmethod
    def _from_string(cls, string):
        try:
            return cls(string)
        except Exception:
            return getattr(cls, string)

    @classmethod
    def strict_from_string(cls, string):
        """Return an instance of FieldEnum instantiated with string."""
        try:
            return cls._from_string(string)
        except (AttributeError, TypeError):
            return None

    @classmethod
    def from_string(cls, string):
        """Return an instance of FieldEnum instantiated with string or an instancxe of UnknownEnumValue if not found."""
        try:
            return cls._from_string(string)
        except (AttributeError, TypeError):
            return UnknownEnumValue(string)


class FuzzyFieldEnum(FieldEnum):
    """A enum representing a field value that can be instantiated with a fuzzy match."""

    @classmethod
    def from_string_ext(cls, string):
        """Return an instance of FieldEnum instantiated with string using a fuzzy match."""
        for name, value in cls.__members__.items():
            if name.lower() in str(string).lower():
                return value
        return UnknownEnumValue(string)

    @classmethod
    def from_string(cls, string):
        """Return an instance of FieldEnum instantiated with string."""
        try:
            return cls._from_string(string)
        except (AttributeError, TypeError):
            return cls.from_string_ext(string)


class Switch(FieldEnum):
    """An enum representing a FIT switch field value."""

    off         = 0
    on          = 1
    auto        = 2
    invalid     = 255


class FitBaseUnit(FieldEnum):
    other       = 0
    kg          = 1
    lb          = 2
    invalid     = 255


class DisplayMeasure(FuzzyFieldEnum):
    metric      = 0
    statute     = 1
    nautical    = 2
    invalid     = 255


class DisplayHeart(FieldEnum):
    """An enum that defines how heart rate will be displayed."""

    bpm     = 0
    max     = 1
    reserve = 2
    invalid = 255


class DisplayPosition(FieldEnum):
    """An enum that defines how position data will be displayed."""

    degree = 0
    dregree_minute = 1
    degree_minute_second = 2
    australian_grid = 3
    british_grid = 4
    dutch_grid = 5
    hugarian_grid = 6
    finish_grid = 7
    german_grid = 8
    icelandic_grid = 9
    indonesian_equatorial = 10
    indonesian_irian = 11
    indonesian_southern = 12
    india_zone_0 = 13
    india_zone_ia = 14
    india_zone_ib = 15
    india_zone_iia = 16
    india_zone_iib = 17
    india_zone_iiia = 18
    india_zone_iiib = 19
    india_zone_iva = 20
    india_zone_ivb = 21
    irish_traverse = 22
    irish_grid = 23
    loran = 24
    maidenhead_grid = 25
    mgrs_grid = 26
    new_zealand_grid = 27
    new_zealand_traverse = 28
    qatar_grid = 29
    modified_swedish_grid = 30
    swedish_grid = 31
    south_african_grid = 32
    swiss_grid = 33
    tiawan_grid = 34
    united_stated_grid = 35
    utm_ups_grid = 36
    west_malayan = 37
    borneo_rso = 38
    estonian_grid = 39
    latvian_grid = 40
    swedish_ref_99_grid = 41
    invalid = 255


class DisplayOrientation(FieldEnum):
    auto = 0
    portrait = 1
    landscape = 2
    portrait_flipped = 3
    landscape_flipped = 4
    invalid = 255


class Side(FieldEnum):
    """An enum that defines what side of the body a device is on."""

    right = 0
    left = 1
    invalid = 255


class BacklightMode(FieldEnum):
    off = 0
    manual = 1
    key_and_messages = 2
    auto_brightness = 3
    smart_notifications = 4
    key_and_messages_night = 5
    key_and_messages_and_smart_notifications = 6
    invalid = 255


class AntNetwork(FieldEnum):
    public  = 0
    antplus = 1
    antfs   = 2
    invalid = 255


class SourceType(FieldEnum):
    ant = 0
    antplus = 1
    bluetooth = 2
    bluetooth_low_energy = 3
    wifi = 4
    local = 5
    invalid = 255


class BatteryStatus(FieldEnum):
    new = 1
    good = 2
    ok = 3
    low = 4
    critical = 5
    charging = 6
    unknown = 7
    invalid = 255


class AutoSyncFrequency(FieldEnum):
    never = 0
    occasionally = 1
    frequent = 2
    once_a_day = 3
    remote = 4
    invalid = 255


class BodyLocation(FieldEnum):
    left_leg = 0
    left_calf = 1
    left_shin = 2
    left_hamstring = 3
    left_quad = 4
    left_glute = 5
    right_leg = 6
    right_calf = 7
    right_shin = 8
    right_hamstring = 9
    right_quad = 10
    right_glut = 11
    torso_back = 12
    left_lower_back = 13
    left_upper_back = 14
    right_lower_back = 15
    right_upper_back = 16
    torso_front = 17
    left_abdomen = 18
    left_chest = 19
    right_abdomen = 20
    right_chest = 21
    left_arm = 22
    left_shoulder = 23
    left_bicep = 24
    left_tricep = 25
    left_brachioradialis = 26
    left_forearm_extensors = 27
    right_arm = 28
    right_shoulder = 29
    right_bicep = 30
    right_tricep = 31
    right_brachioradialis = 32
    right_forearm_extensors = 33
    neck = 34
    throat = 35
    waist_mid_back = 36
    waist_front = 37
    waist_left = 38
    waist_right = 39
    invalid = 255


class Gender(FieldEnum):
    female = 0
    male = 1


class HeartRateZoneCalc(FieldEnum):
    custom = 0
    percent_max_hr = 1
    percent_hrr = 2
    invalid = 255


class PowerCalc(FieldEnum):
    custom = 0
    percent_ftp = 1
    invalid = 255


class Language(FieldEnum):
    English = 0
    French = 1
    Italian = 2
    German = 3
    Spanish = 4
    Croation = 5
    Czech = 6
    Danish = 7
    Dutch = 8
    Finnish = 9
    Greek = 10
    Hungarian = 11
    Norwegian = 12
    Polish = 13
    Portuguese = 14
    Slovakian = 15
    Slovenian = 16
    Swedish = 17
    Russian = 18
    Turkish = 19
    Latvian = 20
    Ukranian = 21
    Arabic = 22
    Farsi = 23
    Bulgarian = 24
    Romanian = 25
    Chinese = 26
    Japanese = 27
    Korean = 28
    Taiwanese = 29
    Thai = 30
    Hebrew = 31
    Brazialn_Portuguese = 32
    Indonesian = 33
    Maylasian = 34
    Vietnamese = 35
    Burmese = 36
    Mongolian = 37
    Custom = 254
    Invalid = 255


class DateMode(FieldEnum):
    day_month   = 0
    month_day   = 1
    invalid     = 255


class TimeMode(FieldEnum):
    twelve_hour             = 0
    twentyfour_hour         = 1
    military                = 2
    twelve_hour_secs        = 3
    twentyfour_hour_secs    = 4
    utc                     = 5
    invalid                 = 255


class Activity(FieldEnum):
    manual              = 0
    auto_multi_sport    = 1


class ActivityType(FieldEnum):
    generic             = 0
    running             = 1
    cycling             = 2
    transition          = 3
    fitness_equipment   = 4
    swimming            = 5
    walking             = 6
    sedentary           = 7
    stop_disable        = 8
    unknown             = 9
    all                 = 245
    invalid             = 255


class Event(FieldEnum):
    timer = 0
    workout = 3
    workout_step = 4
    power_down = 5
    power_up = 6
    off_course = 7
    session = 8
    lap = 9
    course_point = 10
    battery = 11
    virtual_partner_pace = 12
    hr_high_alert = 13
    hr_low_alert = 14
    speed_high_alert = 15
    speed_low_alert = 16
    cad_high_alert = 17
    cad_low_alert = 18
    power_high_alert = 19
    power_low_alert = 20
    recovery_hr = 21
    battery_low = 22
    time_duration_alert = 23
    distance_duration_alert = 24
    calorie_duration_alert = 25
    activity = 26
    fitness_equipment = 27
    length = 28
    user_marker = 32
    sport_point = 33
    calibration = 36
    unknown = 41
    front_gear_change = 42
    rear_gear_change = 43
    rider_position_change = 44
    elev_high_alert = 45
    elev_low_alert = 46
    comm_timeout = 47


class EventType(FieldEnum):
    start = 0
    stop = 1
    consecutive_depreciated = 2
    marker = 3
    stop_all = 4
    begin_depreciated = 5
    end_depreciated = 6
    end_all_depreciated = 7
    stop_disable = 8
    stop_disable_all = 9


class LapTrigger(FieldEnum):
    manual = 0
    time = 1
    distance = 2
    position_start = 4
    position_waypoint = 5
    position_marked = 6
    session_end = 7
    fitness_equipment = 8


class SessionTrigger(FieldEnum):
    activity_end = 0
    manual = 1
    auto_multi_sport = 2
    fitness_equipment = 3


class PersonalRecordType(FieldEnum):
    time        = 0
    distance    = 1
    elevation   = 2
    power       = 3


class GoalType(FieldEnum):
    time            = 0
    distance        = 1
    calories        = 2
    frequency       = 3
    steps           = 4
    ascent          = 5
    active_minutes  = 6


class GoalRecurrence(FieldEnum):
    off             = 0
    daily           = 1
    weekly          = 2
    monthly         = 3
    yearly          = 4
    custom          = 5


class GoalSource(FieldEnum):
    auto            = 0
    community       = 1
    user            = 2


class WatchFaceMode(FieldEnum):
    """A enum that describes a mode of the watch face."""

    digital         = 0
    analog          = 1
    connect_iq      = 2
    disabled        = 3


class ClimbProEvent(FieldEnum):
    """A enum that contains an event from a climbing program."""

    approach        = 0
    start           = 1
    complete        = 2


class HeartRateZonesTimerType(FieldEnum):
    """Gives the type of the Heart Rate Zone Timer."""

    session = 18
    lap = 19


class HeartRateZonesMethod(FieldEnum):
    """Gives the method used to calculate the heart rate zones."""

    max_heart_rate = 1
    heart_rate_reserve = 2
    lactate_threshhold = 3



"""Structured data for decoding a FIT file definition message."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

from .message_type import MessageType
from .fields import Field, TimestampField, TimeMsField, TimeSField, TimeMinField, TimeHourField, TimeOffsetField, TimeOfDayField, NamedField, PowerField, PercentField, \
    TrainingEffectField, CaloriesField, LeftRightBalanceField, WorkField, FractionalCadenceField, FractionalCyclesField, BytePercentField, MessageIndexField, VersionField, \
    CyclesDistanceField, CyclesCaloriesField, CaloriesDayField
from .type_fields import IntegerField, FloatField, StringField, BoolField, HeartRateField, RespirationRateField, AbsolutePressureField, BytesField, MetaMaxField, FitBaseTypeField
from .enum_fields import FileTypeField, TimeModeField, SwitchField, DateModeField, DisplayMeasureField, DisplayHeartField, DisplayPositionField, HeartRateZoneCalcField, \
    PowerCalcField, EventField, EventTypeField, SessionTriggerField, LapTriggerField, FitBaseUnitField, HeartRateZonesTimerTypeField, HeartRateZonesMethodField, ClimbProEventField
from .object_fields import HeightField, WeightField, EnhancedDistanceMetersField, CompressedSpeedDistanceField, DistanceMillimetersField, DistanceCentimetersToMetersField, \
    DistanceMillimetersToMetersField, DistanceCentimetersToKmsField, DistanceMetersField, SpeedMpsField, LatiitudeField, LongitudeField, AltitudeField, TemperatureField
from .activity_fields import ActivityClassField, SportField, SubSportField, SportBasedCyclesField, SportBasedCadenceField, CadenceField, ActivityBasedCyclesField, \
    ActivityTypeField, ActivityTypeIntensityField, ActivityField
from .record_fields import GoalTypeField, GoalValueField, GoalRecurrenceField, GoalSourceField, PersonalRecordTypeField, PersonalRecordField
from .manufacturer_product_fields import ManufacturerField, ProductField
from .device_fields import BacklightModeField, DisplayOrientationField, SideField, AutoSyncFrequencyField, AutoActivityDetectField, DeviceTypeField, BatteryVoltageField, \
    BatteryStatusField, BodyLocationField, AntNetworkField, SourceTypeField, WatchFaceModeField
from .user_fields import GenderField, LanguageField


class DefinitionMessageData():
    """Structured data for decoding a FIT file definition message."""

    known_messages = {
        MessageType.file_id : {
            0 : FileTypeField('type'),
            1 : ManufacturerField(),
            2 : ProductField(),
            3 : IntegerField('serial_number'),
            4 : TimestampField('time_created', utc=True),
            5 : IntegerField('number'),
            7 : StringField('product_name')
        },
        MessageType.capabilities : {},
        MessageType.device_settings : {
            0 : NamedField('active_time_zone'),
            1 : FloatField('utc_offset'),
            2 : TimeOffsetField(),
            4 : TimeModeField(),
            5 : TimeHourField('time_zone_offset'),
            12 : BacklightModeField(),
            35 : SwitchField('switch_35'),
            36 : BoolField('activity_tracker_enabled'),
            38 : SwitchField('switch_38'),
            39 : TimestampField('clock_time', utc=True),
            40 : NamedField('pages_enabled'),
            41 : SwitchField('switch_41'),
            42 : SwitchField('switch_42'),
            43 : SwitchField('switch_43'),
            44 : SwitchField('switch_44'),
            45 : SwitchField('switch_45'),
            46 : BoolField('move_alert_enabled'),
            47 : DateModeField(),
            48 : SwitchField('switch_48'),
            49 : SwitchField('switch_49'),
            52 : SwitchField('switch_52'),
            53 : SwitchField('switch_53'),
            55 : DisplayOrientationField('display_orientation'),
            56 : SideField(),
            57 : NamedField('default_page'),
            58 : NamedField('autosync_min_steps'),
            59 : NamedField('autosync_min_time'),
            64 : SwitchField('switch_64'),
            65 : SwitchField('switch_65'),
            68 : SwitchField('switch_68'),
            80 : BoolField('lactate_threshold_autodetect_enabled'),
            81 : SwitchField('switch_81'),
            82 : SwitchField('switch_82'),
            83 : SwitchField('switch_83'),
            84 : SwitchField('switch_84'),
            85 : SwitchField('switch_85'),
            86 : BoolField('ble_auto_upload_enabled'),
            87 : SwitchField('switch_87'),
            89 : AutoSyncFrequencyField(),
            90 : AutoActivityDetectField(),
            94 : IntegerField('number_of_screens'),
            95 : DisplayOrientationField('smart_notification_display_orientation'),
            107 : SwitchField('switch_107'),
            108 : SwitchField('switch_108'),
            109 : SwitchField('switch_109'),
            110 : SwitchField('switch_110'),
            111 : SwitchField('switch_111'),
            112 : SwitchField('switch_112'),
            126 : SwitchField('switch_126'),
            127 : SwitchField('switch_127'),
            128 : SwitchField('switch_128'),
            133 : SwitchField('switch_133'),
            134 : SwitchField('tap_interface'),
            141 : SwitchField('switch_141'),
        },
        MessageType.user_profile : {
            1 : GenderField(),
            2 : IntegerField('age'),
            3 : HeightField(),
            4 : WeightField('weight'),
            5 : LanguageField(),
            6 : DisplayMeasureField('elev_setting'),
            7 : DisplayMeasureField('weight_setting'),
            8 : HeartRateField('resting_heart_rate'),
            9 : HeartRateField('default_max_running_heart_rate'),
            10 : HeartRateField('default_max_biking_heart_rate'),
            11 : HeartRateField('default_max_heart_rate'),
            12 : DisplayHeartField('hr_setting'),
            13 : DisplayMeasureField('speed_setting'),
            14 : DisplayMeasureField('dist_setting'),
            16 : PowerField('power_setting'),
            17 : ActivityClassField(),
            18 : DisplayPositionField(),
            21 : DisplayMeasureField('temperature_setting'),
            22 : NamedField('local_id'),
            23 : NamedField('global_id'),
            28 : TimeOfDayField('wake_time'),
            29 : TimeOfDayField('sleep_time'),
            30 : DisplayMeasureField('height_setting'),
            31 : EnhancedDistanceMetersField('user_running_step_length'),
            32 : EnhancedDistanceMetersField('user_walking_step_length'),
            35 : TimestampField('ts_35', utc=True),
            41 : TimestampField('ts_41', utc=True),
            47 : DisplayMeasureField('depth_setting'),
            49 : IntegerField('dive_count'),
        },
        MessageType.hrm_profile : {
            0 : BoolField('enabled'),
            1 : NamedField('hrm_ant_id'),
            2 : BoolField('log_hrv'),
            3 : NamedField('hrm_ant_id_trans_type'),
        },
        MessageType.sdm_profile : {},
        MessageType.bike_profile : {
            0 : StringField('name'),
            1 : SportField(),
            2 : SubSportField(),
            3 : DistanceCentimetersToMetersField('odometer'),
            4 : NamedField('bike_spd_ant_id'),
            5 : NamedField('bike_cad_ant_id'),
            6 : NamedField('bike_spdcad_ant_id'),
            7 : NamedField('bike_power_ant_id'),
            8 : DistanceMillimetersToMetersField('custom_wheelsize'),
            9 : DistanceMillimetersToMetersField('auto_wheelsize'),
            10 : WeightField('bike_weight'),
            11 : PercentField('power_cal_factor'),
            12 : BoolField('auto_wheel_cal'),
            13 : BoolField('auto_power_zero'),
            14 : IntegerField('id'),
            15 : BoolField('spd_enabled'),
            16 : BoolField('cad_enabled'),
            17 : BoolField('spdcad_enabled'),
            18 : BoolField('power_enabled'),
            19 : DistanceMillimetersToMetersField('crank_length'),
            20 : BoolField('enabled'),
            21 : NamedField('bike_spd_ant_id_trans_type'),
            22 : NamedField('bike_cad_ant_id_trans_type'),
            23 : NamedField('bike_spdcad_ant_id_trans_type'),
            24 : NamedField('bike_power_ant_id_trans_type'),
            37 : NamedField('odometer_rollover'),
        },
        MessageType.zones_target : {
            1 : HeartRateField('max_heart_rate'),
            2 : HeartRateField('threshold_heart_rate'),
            3 : NamedField('functional_threshold_power'),
            5 : HeartRateZoneCalcField(),
            7 : PowerCalcField()
        },
        MessageType.hr_zone : {
            1 : HeartRateField('high_bpm'),
            2 : StringField('name'),
        },
        MessageType.power_zone : {
            1 : PowerField('high_value'),
            2 : StringField('name'),
        },
        MessageType.met_zone : {},
        MessageType.sport : {
            0 : SportField(),
            1 : SubSportField(),
            3 : StringField('name'),
        },
        MessageType.unknown_13 : {
            5 : LatiitudeField('lat_5'),
            6 : LongitudeField('long_6'),
        },
        MessageType.unknown_14 : {
            10 : Field(name='sequence'),
        },
        MessageType.goal : {
            0 : SportField(),
            1 : SubSportField(),
            2 : TimestampField('start_time', utc=True),
            3 : TimestampField('end_time', utc=True),
            4 : GoalTypeField('type'),
            5 : NamedField('value'),
            6 : BoolField('repeat'),
            7 : GoalValueField(),
            8 : GoalRecurrenceField('recurrence'),
            9 : NamedField('recurrence_value'),
            10 : BoolField('enabled'),
            11 : GoalSourceField('source')
        },
        MessageType.session : {
            0 : EventField(),
            1 : EventField(),
            2 : TimestampField('start_time', utc=True),
            3 : LatiitudeField('start_position_lat'),
            4 : LongitudeField('start_position_long'),
            5 : SportField(),
            6 : SubSportField(),
            7 : TimeMsField('total_elapsed_time'),
            8 : TimeMsField('total_timer_time'),
            9 : DistanceCentimetersToKmsField('total_distance'),
            10 : SportBasedCyclesField('total_cycles'),
            11 : CaloriesField('total_calories'),
            13 : CaloriesField('total_fat_calories'),
            14 : SpeedMpsField('avg_speed'),  # 16 bit version
            15 : SpeedMpsField('max_speed'),  # 16 bit version
            16 : HeartRateField('avg_heart_rate'),
            17 : HeartRateField('max_heart_rate'),
            18 : SportBasedCadenceField('avg_cadence'),
            19 : SportBasedCadenceField('max_cadence'),
            20 : PowerField('avg_power'),
            21 : PowerField('max_power'),
            22 : DistanceMetersField('total_ascent'),
            23 : DistanceMetersField('total_descent'),
            24 : TrainingEffectField('total_training_effect'),
            25 : IntegerField('first_lap_index'),
            26 : IntegerField('num_laps'),
            27 : NamedField('event_group'),
            28 : SessionTriggerField(),
            29 : LatiitudeField('nec_lat'),
            30 : LongitudeField('nec_long'),
            31 : LatiitudeField('swc_lat'),
            32 : LongitudeField('swc_long'),
            33 : IntegerField('length_count'),
            34 : PowerField('normalized_power'),
            35 : TrainingEffectField('training_stress_score'),
            36 : NamedField('intensity_factor'),
            37 : LeftRightBalanceField('left_right_balance'),
            38 : LatiitudeField('end_position_lat'),
            39 : LongitudeField('end_position_long'),
            41 : IntegerField('avg_stroke_count'),
            42 : DistanceCentimetersToMetersField('avg_stroke_distance'),
            43 : NamedField('swim_stroke'),
            44 : DistanceMetersField('pool_length'),
            45 : PowerField('threshold_power'),
            46 : DisplayMeasureField('pool_length_unit'),
            47 : IntegerField('num_active_lengths'),
            48 : WorkField(),
            49 : AltitudeField('avg_altitude'),  # 16 bit version
            50 : AltitudeField('max_altitude'),  # 16 bit version
            51 : DistanceMetersField('gps_accuracy'),
            52 : PercentField('avg_grade'),
            53 : PercentField('avg_pos_grade'),
            54 : PercentField('avg_neg_grade'),
            55 : PercentField('max_pos_grade'),
            56 : PercentField('max_neg_grade'),
            57 : TemperatureField('avg_temperature'),
            58 : TemperatureField('max_temperature'),
            59 : TimeMsField('total_moving_time'),
            60 : SpeedMpsField('avg_pos_vertical_speed'),
            61 : SpeedMpsField('avg_neg_vertical_speed'),
            62 : SpeedMpsField('max_pos_vertical_speed'),
            63 : SpeedMpsField('max_neg_vertical_speed'),
            64 : HeartRateField('min_heart_rate'),
            65 : TimeMsField('time_in_hr_zone'),
            66 : TimeMsField('time_in_speed_zone'),
            67 : TimeMsField('time_in_cadence_zone'),
            68 : TimeMsField('time_in_power_zone'),
            69 : TimeMsField('avg_lap_time'),
            70 : IntegerField('best_lap_index'),
            71 : AltitudeField('min_altitude'),  # 16 bit version
            78 : NamedField('total_swim_time'),
            79 : NamedField('average_stroke'),
            80 : NamedField('swolf'),
            82 : NamedField('player_score'),
            83 : NamedField('opponent_score'),
            84 : StringField('opponent_name'),
            85 : IntegerField('stroke_count'),
            86 : IntegerField('zone_count'),
            87 : SpeedMpsField('max_ball_speed'),
            88 : SpeedMpsField('avg_ball_speed'),
            89 : DistanceMillimetersField('avg_vertical_oscillation'),
            90 : PercentField('avg_stance_time_percent'),
            91 : TimeMsField('avg_stance_time', scale=10.0),
            92 : FractionalCadenceField('avg_fractional_cadence'),
            93 : FractionalCadenceField('max_fractional_cadence'),
            94 : FractionalCyclesField(),
            95 : FloatField('avg_total_hemoglobin_conc'),
            96 : FloatField('min_total_hemoglobin_conc'),
            97 : FloatField('max_total_hemoglobin_conc'),
            98 : PercentField('avg_saturated_hemoglobin_percent'),
            99 : PercentField('min_saturated_hemoglobin_percent'),
            100 : PercentField('max_saturated_hemoglobin_percent'),
            101 : BytePercentField('avg_left_torque_effectiveness'),
            102 : BytePercentField('avg_right_torque_effectiveness'),
            103 : BytePercentField('avg_left_pedal_smoothness'),
            104 : BytePercentField('avg_right_pedal_smoothness'),
            105 : BytePercentField('avg_combined_pedal_smoothness'),
            110 : StringField('sport_name'),
            111 : IntegerField('sport_index'),
            112 : TimeMsField('time_standing'),
            113 : NamedField('stand_count'),
            114 : NamedField('avg_left_pco'),
            115 : NamedField('avg_right_pco'),
            116 : NamedField('avg_left_power_phase'),
            117 : NamedField('avg_left_power_phase_peak'),
            118 : NamedField('avg_right_power_phase'),
            119 : NamedField('avg_right_power_phase_peak'),
            120 : PowerField('avg_power_position'),
            121 : PowerField('max_power_position'),
            122 : CadenceField('avg_cadence_position'),
            123 : CadenceField('max_cadence_position'),
            124 : SpeedMpsField('avg_speed'),  # 32 bit version
            125 : SpeedMpsField('max_speed'),  # 32 bit version
            126 : AltitudeField('avg_altitude'),  # 32 bit version
            127 : AltitudeField('min_altitude'),  # 32 bit version
            128 : AltitudeField('max_altitude'),  # 32 bit version
            129 : PowerField('avg_lev_motor_power'),
            130 : PowerField('max_lev_motor_power'),
            131 : BytePercentField('lev_battery_consumption'),
            132 : PercentField('avg_vertical_ratio'),
            133 : PercentField('avg_stance_time_balance'),
            134 : DistanceMillimetersField('avg_step_length'),
            137 : TrainingEffectField('total_anaerobic_training_effect'),
            139 : SpeedMpsField('avg_vam'),
            150 : TemperatureField('min_temperature'),
            169 : RespirationRateField('avg_respiration_rate'),
            170 : RespirationRateField('max_respiration_rate'),
            178: FloatField('fluid_consumed'),
            180 : RespirationRateField('min_respiration_rate'),
        },
        MessageType.lap : {
            0 : EventField(),
            1 : EventTypeField(),
            2 : TimestampField('start_time', utc=True),
            3 : LatiitudeField('start_position_lat'),
            4 : LongitudeField('start_position_long'),
            5 : LatiitudeField('end_position_lat'),
            6 : LongitudeField('end_position_long'),
            7 : TimeMsField('total_elapsed_time'),
            8 : TimeMsField('total_timer_time'),
            9 : DistanceCentimetersToKmsField('total_distance'),
            10 : SportBasedCyclesField('total_cycles'),
            11 : CaloriesField('total_calories'),
            12 : CaloriesField('total_fat_calories'),
            13 : SpeedMpsField('avg_speed'),  # 16 bit version
            14 : SpeedMpsField('max_speed'),  # 16 bit version
            15 : HeartRateField('avg_heart_rate'),
            16 : HeartRateField('max_heart_rate'),
            17 : SportBasedCadenceField('avg_cadence'),
            18 : SportBasedCadenceField('max_cadence'),
            19 : PowerField('avg_power'),
            20 : PowerField('max_power'),
            21 : DistanceMetersField('total_ascent'),
            22 : DistanceMetersField('total_descent'),
            24 : LapTriggerField(),
            25 : SportField(),
            26 : NamedField('event_group'),
            27 : LatiitudeField('unknown_lat_27'),
            28 : LongitudeField('unknown_long_28'),
            29 : LatiitudeField('unknown_lat_29'),
            30 : LongitudeField('unknown_long_30'),
            32 : IntegerField('num_lengths'),
            33 : PowerField('normalized_power'),
            34 : LeftRightBalanceField('left_right_balance'),
            35 : IntegerField('first_length_index'),
            37 : DistanceCentimetersToMetersField('avg_stroke_distance'),
            38 : NamedField('swim_stroke'),
            39 : SubSportField(),
            40 : IntegerField('num_active_lengths'),
            41 : WorkField(),
            42 : AltitudeField('avg_altitude'),  # 16 bit version
            43 : AltitudeField('max_altitude'),  # 16 bit version
            44 : DistanceMetersField('gps_accuracy'),
            45 : PercentField('avg_grade'),
            46 : PercentField('avg_pos_grade'),
            47 : PercentField('avg_neg_grade'),
            48 : PercentField('max_pos_grade'),
            49 : PercentField('max_neg_grade'),
            50 : TemperatureField('avg_temperature'),
            51 : TemperatureField('max_temperature'),
            52 : TimeMsField('total_moving_time'),
            53 : SpeedMpsField('avg_pos_vertical_speed'),
            54 : SpeedMpsField('avg_neg_vertical_speed'),
            55 : SpeedMpsField('max_pos_vertical_speed'),
            56 : SpeedMpsField('max_neg_vertical_speed'),
            57 : TimeMsField('time_in_hr_zone'),
            58 : TimeMsField('time_in_speed_zone'),
            59 : TimeMsField('time_in_cadence_zone'),
            60 : TimeMsField('time_in_power_zone'),
            61 : IntegerField('repetition_num'),
            62 : AltitudeField('min_altitude'),  # 16 bit version
            63 : HeartRateField('min_heart_rate'),
            70 : NamedField('swim_time'),
            71 : MessageIndexField('wkt_step_index'),
            72 : NamedField('average_stroke'),
            73 : NamedField('swolf'),
            74 : NamedField('opponent_score'),
            75 : IntegerField('stroke_count'),
            76 : IntegerField('zone_count'),
            77 : DistanceMillimetersField('avg_vertical_oscillation'),
            78 : PercentField('avg_stance_time_percent'),
            79 : TimeMsField('avg_stance_time', scale=10.0),
            80 : FractionalCadenceField('avg_fractional_cadence'),
            81 : FractionalCadenceField('max_fractional_cadence'),
            82 : FractionalCyclesField(),
            83 : NamedField('player_score'),
            84 : NamedField('avg_total_hemoglobin_conc'),
            85 : NamedField('min_total_hemoglobin_conc'),
            86 : NamedField('max_total_hemoglobin_conc'),
            87 : PercentField('avg_saturated_hemoglobin_percent'),
            88 : PercentField('min_saturated_hemoglobin_percent'),
            89 : PercentField('max_saturated_hemoglobin_percent'),
            91 : BytePercentField('avg_left_torque_effectiveness'),
            92 : BytePercentField('avg_right_torque_effectiveness'),
            93 : BytePercentField('avg_left_pedal_smoothness'),
            94 : BytePercentField('avg_right_pedal_smoothness'),
            95 : BytePercentField('avg_combined_pedal_smoothness'),
            98 : TimeMsField('time_standing'),
            99 : IntegerField('stand_count'),
            100 : NamedField('avg_left_pco'),
            101 : NamedField('avg_right_pco'),
            102 : NamedField('avg_left_power_phase'),
            103 : NamedField('avg_left_power_phase_peak'),
            104 : NamedField('avg_right_power_phase'),
            105 : NamedField('avg_right_power_phase_peak'),
            106 : PowerField('avg_power_position'),
            107 : PowerField('max_power_position'),
            108 : CadenceField('avg_cadence_position'),
            109 : CadenceField('max_cadence_position'),
            110 : SpeedMpsField('avg_speed'),  # 32 bit version
            111 : SpeedMpsField('max_speed'),  # 32 bit version
            112 : AltitudeField('avg_altitude'),  # 32 bit version
            113 : AltitudeField('min_altitude'),  # 32 bit version
            114 : AltitudeField('max_altitude'),  # 32 bit version
            115 : PowerField('avg_lev_motor_power'),
            116 : PowerField('max_lev_motor_power'),
            117 : BytePercentField('lev_battery_consumption'),
            118 : PercentField('avg_vertical_ratio', 100.0),
            119 : PercentField('avg_stance_time_balance'),
            120 : DistanceMillimetersField('avg_step_length'),
            121 : SpeedMpsField('avg_vam'),
            136 : RespirationRateField('avg_respiration_rate'),
            137 : RespirationRateField('max_respiration_rate'),
        },
        MessageType.record : {
            0 : LatiitudeField('position_lat'),
            1 : LongitudeField('position_long'),
            2 : AltitudeField('altitude'),  # 16 bit version
            3 : HeartRateField('heart_rate'),
            4 : CadenceField(),
            5 : DistanceCentimetersToKmsField('distance'),
            6 : SpeedMpsField('speed'),  # 16 bit version
            7 : PowerField(),
            8 : CompressedSpeedDistanceField(),
            9 : PercentField('grade'),
            10 : NamedField('resistance'),
            11 : TimeMsField('time_from_course'),
            12 : DistanceMetersField('cycle_length', scale=100),
            13 : TemperatureField('temperature'),
            17 : SpeedMpsField('speed_1s'),
            18 : ActivityBasedCyclesField(),
            19 : ActivityBasedCyclesField('total_cycles'),
            28 : NamedField('accumulated_power'),  # 16 bit version
            29 : NamedField('accumulated_power'),  # 32 bit version
            30 : LeftRightBalanceField('left_right_balance'),
            31 : DistanceMetersField('gps_accuracy'),
            32 : SpeedMpsField('vertical_speed'),
            33 : CaloriesField(),
            39 : DistanceMillimetersField('avg_vertical_oscillation'),
            40 : PercentField('stance_time_percent'),
            41 : TimeMsField('stance_time', scale=10.0),
            42 : ActivityTypeField(),
            43 : BytePercentField('left_torque_effectiveness'),
            44 : BytePercentField('right_torque_effectiveness'),
            45 : BytePercentField('left_pedal_smoothness'),
            46 : BytePercentField('right_pedal_smoothness'),
            47 : BytePercentField('combined_pedal_smoothness'),
            48 : NamedField('time128'),
            49 : NamedField('stroke_type'),
            50 : NamedField('zone'),
            51 : NamedField('ball_speed'),
            52 : CadenceField('cadence256'),
            53 : FractionalCadenceField(),
            54 : NamedField('total_hemoglobin_conc'),
            55 : NamedField('total_hemoglobin_conc_min'),
            56 : NamedField('total_hemoglobin_conc_max'),
            57 : PercentField('saturated_hemoglobin_percent'),
            58 : PercentField('saturated_hemoglobin_percent_min'),
            59 : PercentField('saturated_hemoglobin_percent_max'),
            62 : IntegerField('device_index'),
            67 : NamedField('left_pco'),
            68 : NamedField('right_pco'),
            69 : NamedField('left_power_phase'),
            70 : NamedField('left_power_phase_peak'),
            71 : NamedField('right_power_phase'),
            72 : NamedField('right_power_phase_peak'),
            73 : SpeedMpsField('speed'),  # 32 bit version
            78 : AltitudeField('altitude'),  # 32 bit version
            83 : PercentField('vertical_ratio', 100.0),
            84 : PercentField('stance_time_balance'),
            85 : DistanceMillimetersField('step_length'),
            91 : AbsolutePressureField(),
            92 : DistanceMillimetersField('depth'),
            93 : DistanceMillimetersField('next_stop_depth'),
            94 : TimeSField('next_stop_time'),
            95 : TimeSField('time_to_surface'),
            96 : TimeSField('ndl_time'),
            97 : PercentField('cns_load'),
            98 : PercentField('n2_load'),
            108 : RespirationRateField(),
            114 : FloatField('grit'),
            115 : FloatField('flow'),
        },
        MessageType.event : {
            0 : EventField('event'),
            1 : EventTypeField(),
            2 : NamedField('data16'),
            3 : NamedField('data'),
            4 : NamedField('event_group'),
            7 : NamedField('score'),
            8 : NamedField('opponent_score'),
            9 : NamedField('front_gear_num'),
            10 : NamedField('front_gear'),
            11 : NamedField('rear_gear_num'),
            12 : NamedField('rear_gear'),
            13 : IntegerField('device_index'),
            15 : TimestampField('ts_15', utc=True),
        },
        MessageType.source : {},
        MessageType.device_info : {
            0 : IntegerField('device_index'),
            1 : DeviceTypeField(),
            2 : ManufacturerField(),
            3 : IntegerField('serial_number'),
            4 : ProductField(),
            5 : VersionField('software_version'),
            6 : NamedField('hardware_version'),
            7 : TimeMsField('cum_operating_time', scale=1000.0),
            10 : BatteryVoltageField(),
            11 : BatteryStatusField(),
            15 : NamedField('ant_related'),    # only found on ant devices?
            18 : BodyLocationField('sensor_position'),
            19 : StringField('descriptor'),
            20 : NamedField('ant_transmission_type'),
            21 : NamedField('ant_device_number'),
            22 : AntNetworkField(),
            24 : IntegerField('sensor_id'),
            25 : SourceTypeField(),
            27 : StringField('product_name'),
        },
        MessageType.unknown_24 : {
            2 : BytesField('data'),
        },
        MessageType.workout : {
            6 : IntegerField('num_valid_steps'),
            8 : StringField('wkt_name'),
        },
        MessageType.workout_step : {},
        MessageType.schedule : {},
        MessageType.location : {},
        MessageType.weight_scale : {
            0 : WeightField('weight'),
            1 : PercentField('percent_fat'),
            2 : PercentField('percent_hydration'),
            3 : WeightField('visceral_fat_mass'),
            4 : WeightField('bone_mass'),
            5 : WeightField('muscle_mass'),
            7 : NamedField('basal_met'),
            8 : NamedField('physique_rating'),
            9 : NamedField('active_met'),
            10 : NamedField('metabolic_age'),
            11 : NamedField('visceral_fat_rating'),
            12 : NamedField('user_profile_index')
        },
        MessageType.course : {},
        MessageType.course_point : {},
        MessageType.totals : {
            0 : TimeSField('timer_time'),
            1 : DistanceMetersField('distance'),
            2 : CaloriesField(),
            3 : SportField(),
            4 : TimeSField('elapsed_time'),
            5 : IntegerField('sessions'),
            6 : TimeSField('active_time'),
            9 : IntegerField('sport_index'),
        },
        MessageType.activity : {
            0 : TimeMsField('total_timer_time'),
            1 : IntegerField('num_sessions'),
            2 : ActivityField(),
            3 : EventField(),
            4 : EventTypeField(),
            5 : TimestampField('local_timestamp', utc=False),
            6 : NamedField('event_group'),
        },
        MessageType.software : {
            3 : VersionField()
        },
        MessageType.file_capabilities : {},
        MessageType.mesg_capabilities : {},
        MessageType.field_capabilities : {},
        MessageType.file_creator : {
            0 : VersionField('software_version'),
            1 : VersionField('hardware_version'),
            2 : BytesField('data')
        },
        MessageType.blood_pressure : {},
        MessageType.speed_zone : {},
        MessageType.monitoring : {
            0 : IntegerField('device_index'),
            1 : CaloriesField(),
            2 : DistanceCentimetersToMetersField('distance'),
            3 : ActivityBasedCyclesField(),
            4 : TimeMsField('cum_active_time'),
            5 : ActivityTypeField(),
            6 : NamedField('activity_subtype'),
            7 : NamedField('activity_level'),
            14 : TemperatureField('temperature_min'),
            15 : TemperatureField('temperature_max'),
            19 : CaloriesField('active_calories'),
            24 : ActivityTypeIntensityField('current_activity_type_intensity'),
            26 : TimeSField('timestamp_16'),
            27 : HeartRateField('heart_rate'),
            29 : TimeMinField('duration'),
            31 : DistanceMillimetersToMetersField('ascent'),
            32 : DistanceMillimetersToMetersField('descent'),
            33 : TimeMinField('moderate_activity_time'),
            34 : TimeMinField('vigorous_activity_time'),
            35 : DistanceMillimetersToMetersField('cum_ascent'),
            36 : DistanceMillimetersToMetersField('cum_descent')
        },
        MessageType.training_file : {
            0 : FileTypeField('type'),
            1 : ManufacturerField(),
            2 : ProductField(),
            3 : IntegerField('serial_number'),
            4 : TimestampField('time_created', utc=True),
        },  # timestamp, serial_number, creation_time, product_ID, session_style
        MessageType.hrv : {
            0 : TimeMsField('time'),
        },
        MessageType.ant_rx : {},
        MessageType.ant_tx : {},
        MessageType.ant_channel_id : {},
        MessageType.length : {},
        MessageType.monitoring_info : {
            0 : TimestampField('local_timestamp', utc=False),
            1 : ActivityTypeField(),
            3 : CyclesDistanceField(),
            4 : CyclesCaloriesField(),
            5 : CaloriesDayField('resting_metabolic_rate')
        },
        MessageType.battery : {
            0 : TimeMinField('remaining_mins'),
            2 : PercentField('charge'),
            3 : BatteryVoltageField(scale=10.0),
        },
        MessageType.pad : {},
        MessageType.slave_device : {},
        MessageType.personal_record : {
            0 : PersonalRecordTypeField(),
            1 : SportField(),
            2 : DistanceCentimetersToMetersField('record_distance'),
            3 : DistanceCentimetersToMetersField('record_distance2'),
            4 : DistanceCentimetersToMetersField('actual_distance'),
            5 : PersonalRecordField()
        },
        MessageType.connectivity : {
            0 : BoolField('bluetooth_enabled'),
            1 : BoolField('bluetooth_le_enabled'),
            2 : BoolField('ant_enabled'),
            3 : StringField('name'),
            4 : BoolField('live_tracking_enabled'),
            5 : BoolField('weather_conditions_enabled'),
            6 : BoolField('weather_alerts_enabled'),
            7 : BoolField('auto_activity_upload_enabled'),
            8 : BoolField('course_download_enabled'),
            9 : BoolField('workout_download_enabled'),
            10 : BoolField('gps_ephemeris_download_enabled'),
            11 : BoolField('incident_detection_enabled'),
            12 : BoolField('grouptrack_enabled'),
        },
        MessageType.weather_conditions : {},
        MessageType.weather_alert : {},
        MessageType.cadence_zone : {},
        MessageType.hr : {},
        MessageType.firstbeat : {
            1 : HeartRateField('max_heart_rate'),
            4 : TrainingEffectField('aerobic_training_effect'),
            7 : MetaMaxField(),
            9 : TimeMinField('recovery_time'),
            14 : HeartRateField('lactate_threshold_heart_rate'),
            15 : SpeedMpsField('lactate_threshold_speed'),  # 16 bit version
            21 : LatiitudeField('position_21'),
            24 : LongitudeField('position_24'),
        },
        MessageType.unknown_141 : {
            1 : TimestampField('ts_1', utc=False),
            2 : TimestampField('ts_2', utc=False),
            4 : LatiitudeField('position_4'),
            5 : LongitudeField('position_5'),
        },
        MessageType.segment_lap : {},
        MessageType.memo_glob : {},
        MessageType.sensor : {
            0 : IntegerField('sensor_id'),
            2 : StringField('name'),
            10 : DistanceMillimetersToMetersField('manual_wheelsize'),
            11 : FloatField('calibration_factor', scale=10.0),
            21 : DistanceMillimetersToMetersField('auto_wheelsize'),
            32 : ProductField(),
            33 : ManufacturerField(),
            34 : VersionField('software_version'),
            50 : BytesField('ble_id')
        },
        MessageType.segment_id : {},
        MessageType.segment_leaderboard_entry : {},
        MessageType.segment_point : {},
        MessageType.segment_file : {},
        MessageType.workout_session : {},
        MessageType.watchface_settings : {
            0 : WatchFaceModeField('mode'),
            1 : NamedField('layout'),
        },
        MessageType.gps_metadata : {},
        MessageType.camera_event : {},
        MessageType.timestamp_correlation : {},
        MessageType.gyroscope_data : {},
        MessageType.accelerometer_data : {},
        MessageType.three_d_sensor_calibration : {},
        MessageType.video_frame : {},
        MessageType.obdii_data : {},
        MessageType.nmea_sentence : {},
        MessageType.aviation_attitude : {},
        MessageType.video : {},
        MessageType.video_title : {},
        MessageType.video_description : {},
        MessageType.video_clip : {},
        MessageType.ohr_settings : {
            0 : SwitchField('enabled'),
        },
        MessageType.exd_screen_configuration : {},
        MessageType.exd_data_field_configuration : {},
        MessageType.exd_data_concept_configuration : {},
        MessageType.field_description : {
            0 : IntegerField('developer_data_index'),
            1 : NamedField('field_definition_number'),
            2 : FitBaseTypeField('fit_base_type_id'),
            3 : StringField('field_name'),
            4 : NamedField('array'),
            5 : StringField('components'),
            6 : FloatField('scale'),
            7 : FloatField('offset'),
            8 : StringField('units'),
            9 : StringField('bits'),
            10 : StringField('accumulate'),
            13 : FitBaseUnitField('fit_base_unit_id'),
            14 : IntegerField('native_message_num'),
            15 : IntegerField('native_field_num')
        },
        MessageType.dev_data_id : {
            0 : NamedField('developer_id'),
            1 : BytesField('application_id'),
            2 : ManufacturerField(),
            3 : IntegerField('developer_data_index'),
            4 : NamedField('application_version')
        },
        MessageType.magnetometer_data : {},
        MessageType.barometer_data : {},
        MessageType.one_d_sensor_calibration : {},
        MessageType.hr_zones_timer : {
            0 : HeartRateZonesTimerTypeField(),
            1 : IntegerField('record_num'),
            2 : TimeMsField('hr_zones_time'),
            6 : IntegerField('hr_zones'),
            10: HeartRateZonesMethodField(),
            11 : HeartRateField('max_heart_rate'),
            12 : HeartRateField('min_heart_rate'),
            13 : HeartRateField('lactate_threshhold_heart_rate')
        },
        MessageType.set : {},
        MessageType.stress_level : {
            0 : NamedField('stress_level'),
            1 : TimestampField('local_timestamp', utc=False),
        },
        MessageType.unknown_233 : {
            2 : BytesField('unknown_2'),
        },
        MessageType.local_time : {
            0 : TimestampField('local_timestamp', utc=False)
        },
        MessageType.dive_settings : {},
        MessageType.dive_gas : {},
        MessageType.dive_alarm : {},
        MessageType.exercise_title : {},
        MessageType.dive_summary : {},
        MessageType.pulse_ox : {
            0 : FloatField('pulse_ox'),
        },
        # Names and types for this message are guesses
        MessageType.start : {
            2 : TimestampField('local_timestamp', utc=False),
        },
        # Names and types for this message are guesses
        MessageType.sleep_data : {
            0 : BytesField('data'),
        },
        # Names and types for this message are guesses
        MessageType.sleep_level : {
            0 : FloatField('sleep_level'),
        },
        # Names and types for this message are guesses
        MessageType.end : {},
        MessageType.unknown_284 : {
            1 : TimestampField('ts_1', utc=True),
        },
        MessageType.jump : {
            0 : DistanceMetersField('distance'),
            1 : DistanceMetersField('height'),
            2 : IntegerField('rotations'),
            3 : FloatField('hang_time'),
            4 : FloatField('score'),
            5 : LatiitudeField('position_lat'),
            6 : LongitudeField('position_long'),
            7 : SpeedMpsField('speed'),
            8 : SpeedMpsField('speed'),
        },
        MessageType.respiration : {
            0 : RespirationRateField(),
        },
        MessageType.climb_pro : {
            0 : LatiitudeField('position_lat'),
            1 : LongitudeField('position_long'),
            2 : ClimbProEventField('climb_pro_event'),
            3 : IntegerField('climb_number'),
            4 : IntegerField('climb_category'),
            5 : DistanceMetersField('current_dist')
        },
        MessageType.mfg_range_min : {},
        MessageType.mfg_range_max : {},
    }
    reserved_field_indexes = {
        250 : IntegerField('part_index'),
        253 : TimestampField('timestamp', utc=True),
        254 : MessageIndexField('message_index')
    }

    @classmethod
    def get_message_definition(cls, message_type):
        """Given a message number, return a message definition."""
        return cls.known_messages.get(message_type, {})



"""Enums that represent FIT file message manufacturer field values."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

from .field_enums import FuzzyFieldEnum


class Manufacturer(FuzzyFieldEnum):
    """Manufacturer codes used in FIT files."""

    #
    # Garmin defined values
    #
    Garmin                          = 1
    garmin_fr405_antfs              = 2
    zephyr                          = 3
    dayton                          = 4
    idt                             = 5
    srm                             = 6
    quarq                           = 7
    ibike                           = 8
    saris                           = 9
    spark_hk                        = 10
    tanita                          = 11
    echowell                        = 12
    Dynastream_OEM                  = 13
    nautilus                        = 14
    Dynastream                      = 15
    timex                           = 16
    metrigear                       = 17
    xelic                           = 18
    beurer                          = 19
    cardiosport                     = 20
    a_and_d                         = 21
    hmm                             = 22
    Suunto                          = 23
    thita_elektronik                = 24
    gpulse                          = 25
    clean_mobile                    = 26
    pedal_brain                     = 27
    peaksware                       = 28
    saxonar                         = 29
    lemond_fitness                  = 30
    dexcom                          = 31
    Wahoo_Fitness                   = 32
    Octane_Fitness                  = 33
    archinoetics                    = 34
    the_hurt_box                    = 35
    citizen_systems                 = 36
    magellan                        = 37
    osynce                          = 38
    holux                           = 39
    concept2                        = 40
    one_giant_leap                  = 42
    ace_sensor                      = 43
    brim_brothers                   = 44
    xplova                          = 45
    perception_digital              = 46
    bf1systems                      = 47
    pioneer                         = 48
    spantec                         = 49
    metalogics                      = 50
    _4iiiis                         = 51
    seiko_epson                     = 52
    seiko_epson_oem                 = 53
    ifor_powell                     = 54
    maxwell_guider                  = 55
    Star_Trac                       = 56
    breakaway                       = 57
    alatech_technology_ltd          = 58
    mio_technology_europe           = 59
    rotor                           = 60
    geonaute                        = 61
    id_bike                         = 62
    Specialized                     = 63
    wtek                            = 64
    physical_enterprises            = 65
    north_pole_engineering          = 66
    bkool                           = 67
    cateye                          = 68
    stages_cycling                  = 69
    sigmasport                      = 70
    TomTom                          = 71
    peripedal                       = 72
    wattbike                        = 73
    moxy                            = 76
    ciclosport                      = 77
    powerbahn                       = 78
    acorn_projects_aps              = 79
    lifebeam                        = 80
    Bontrager                       = 81
    wellgo                          = 82
    Scosche                         = 83
    magura                          = 84
    woodway                         = 85
    elite                           = 86
    nielsen_kellerman               = 87
    dk_city                         = 88
    tacx                            = 89
    direction_technology            = 90
    magtonic                        = 91
    OnePartCarbon                   = 92
    inside_ride_technologies        = 93
    sound_of_motion                 = 94
    stryd                           = 95
    icg                             = 96
    MiPulse                         = 97
    bsx_athletics                   = 98
    look                            = 99
    campagnolo_srl                  = 100
    body_bike_smart                 = 101
    praxisworks                     = 102
    limits_technology               = 103
    topaction_technology            = 104
    cosinuss                        = 105
    fitcare                         = 106
    magene                          = 107
    giant_manufacturing_co          = 108
    tigrasport                      = 109
    salutron                        = 110
    TechnoGym                       = 111
    Bryton_Sensors                  = 112
    Latitude_Limited                = 113
    Soaring_Technology              = 114
    IGPSport                        = 115
    ThinkRider                      = 116
    GopherSport                     = 117
    WaterPower                      = 118
    OrangeTheory                    = 119
    Inpeak                          = 120
    Kinetic                         = 121
    Johnson_Health_Tech             = 122
    Polar_Electro                   = 123
    Seesense                        = 124
    Garmin_local_154                = 154
    Garmin_local_218                = 218
    development                     = 255
    Health_and_Life                 = 257
    lezyne                          = 258
    scribe_labs                     = 259
    zwift                           = 260
    watteam                         = 261
    recon                           = 262
    favero_electronics              = 263
    dynovelo                        = 264
    Strava                          = 265
    precor                          = 266
    bryton                          = 267
    sram                            = 268
    navman                          = 269
    cobi                            = 270
    spivi                           = 271
    mio_magellan                    = 272
    evesports                       = 273
    sensitivus_gauge                = 274
    podoon                          = 275
    Life_Time_Fitness               = 276
    falco_e_motors                  = 277
    Minoura                         = 278
    Cycliq                          = 279
    Luxottica                       = 280
    Trainer_Road                    = 281
    The_Sufferest                   = 282
    FullSpeedAhead                  = 283
    VirtualTraining                 = 284
    FeedbackSports                  = 285
    Omata                           = 286
    VDO                             = 287
    MagneticDays                    = 288
    Hammerhead                      = 289
    Kinetic_By_Kurt                 = 290
    Shapelog                        = 291
    Dabuziduo                       = 292
    JetBlack                        = 293
    Coros                           = 294
    virtugo                         = 295
    velosense                       = 296
    cycligentinc                    = 297
    trailforks                      = 298
    mahle_ebikemotion               = 299
    #
    nurvv                           = 300
    MicroProgram                    = 301
    Zone5Cloud                      = 302
    GreenTEG                        = 303
    Yamaha_Motors                   = 304
    Whoop                           = 305
    GRAVAA                          = 306
    OneLap                          = 307
    Monark_Exercise                 = 308
    Form                            = 309
    Decathlon                       = 310
    #
    actigraphcorp                   = 5759
    #
    # Privates values
    #
    Garmin_local                    = 0
    Garmin_local_start              = 500
    unknown                         = 65534
    invalid                         = 65535



"""Objects that represent FIT file object message fields."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


from .fields import NamedField
from .field_enums import FitBaseUnit, Switch, DisplayMeasure, DisplayHeart, DisplayPosition, HeartRateZoneCalc, PowerCalc, DateMode, TimeMode, Event, EventType, LapTrigger, \
    SessionTrigger, ClimbProEvent, HeartRateZonesTimerType, HeartRateZonesMethod
from .file_type import FileType


class EnumField(NamedField):
    """Base class for a field that can be represented by an enum value."""

    _enum = None

    def _convert_single(self, value, invalid):
        return self._enum.from_string(value)


class SwitchField(EnumField):
    """A field whose value can be represented by a Switch enum."""

    _enum = Switch


class FitBaseUnitField(EnumField):

    _enum = FitBaseUnit


class DisplayMeasureField(EnumField):

    _enum = DisplayMeasure


class DisplayHeartField(EnumField):

    _enum = DisplayHeart


class DisplayPositionField(EnumField):

    _name = 'position_setting'
    _enum = DisplayPosition


#
# Hardware related fields
#
class HeartRateZoneCalcField(EnumField):

    _name = 'hr_calc_type'
    _enum = HeartRateZoneCalc


class PowerCalcField(EnumField):

    _name = 'pwr_calc_type'
    _enum = PowerCalc


class DateModeField(EnumField):

    _name = 'date_mode'
    _enum = DateMode


class TimeModeField(EnumField):

    _name = 'time_mode'
    _enum = TimeMode


class FileTypeField(EnumField):
    """A field that indicates the FIT file type."""

    _name = 'file_type'
    _enum = FileType


class EventField(EnumField):
    """A field that contains an event."""

    _name = 'event'
    _enum = Event


class EventTypeField(EnumField):
    """A field that contains the type of an event."""

    _name = 'event_type'
    _enum = EventType


class LapTriggerField(EnumField):
    """A field that indicates why a lap was started."""

    _name = 'lap_trigger'
    _enum = LapTrigger


class SessionTriggerField(EnumField):
    """A field that indicates why a session was started."""

    _name = 'session_trigger'
    _enum = SessionTrigger


class ClimbProEventField(EnumField):
    """A field that contains an event from a climbing program."""

    _name = 'climb_pro_event'
    _enum = ClimbProEvent


class HeartRateZonesTimerTypeField(EnumField):
    """A field that contains the type of the Heart Rate Zone Timer."""

    _name = 'hr_zones_timer_type'
    _enum = HeartRateZonesTimerType


class HeartRateZonesMethodField(EnumField):
    """A field that contains the method used to calculate the heart rate zones."""

    _name = 'hr_zones_method'
    _enum = HeartRateZonesMethod



"""Enums that represent FIT file message file type field values."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

from .field_enums import FieldEnum


class FileType(FieldEnum):
    """Garmin defined values."""

    unknown0                    = 0
    device                      = 1
    settings                    = 2
    sport_settings              = 3
    activity                    = 4
    workout                     = 5
    course                      = 6
    schedules                   = 7
    weight                      = 9
    totals                      = 10
    goals                       = 11
    blood_pressure              = 14
    monitoring_a                = 15
    activity_summary            = 20
    monitoring_daily            = 28
    records                     = 29
    monitoring_b                = 32
    segment                     = 34
    segment_list                = 35
    exd_configuration           = 40
    metrics                     = 44
    sleep                       = 49
    unknown_file_type_64        = 64
    manufacturer_range_start    = 0xfe
    invalid                     = 255
